#!/usr/bin/python
# -*- coding: utf-8 -*-
#################################################
### Extracted from platform version: 4.1.2 
#################################################
AOSP_PERMISSIONS = {
	'android.permission.REMOTE_AUDIO_PLAYBACK' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : '', 'protectionLevel' : 'signature', 'label' : ''},
	'android.permission.ACCESS_COARSE_LOCATION' : {'permissionGroup' : 'android.permission-group.LOCATION', 'description' : 'Access approximate location from location providers using network sources such as cell tower and Wi-Fi. When these location services are available and turned on, this permission allows the app to determine your approximate location.', 'protectionLevel' : 'dangerous', 'label' : 'approximate (network-based) location'},
	'android.permission.BIND_WALLPAPER' : {'permissionGroup' : '', 'description' : 'Allows the holder to bind to the top-level interface of a wallpaper. Should never be needed for normal apps.', 'protectionLevel' : 'signature|system', 'label' : 'bind to a wallpaper'},
	'android.permission.FORCE_BACK' : {'permissionGroup' : '', 'description' : 'Allows the app to force any activity that is in the foreground to close and go back. Should never be needed for normal apps.', 'protectionLevel' : 'signature', 'label' : 'force app to close'},
	'android.permission.READ_CALENDAR' : {'permissionGroup' : 'android.permission-group.PERSONAL_INFO', 'description' : 'Allows the app to read all calendar events stored on your phone, including those of friends or co-workers. This may allow the app to share or save your calendar data, regardless of confidentiality or sensitivity.', 'protectionLevel' : 'dangerous', 'label' : 'read calendar events plus confidential information'},
	'ti.permission.FMRX' : {'permissionGroup' : 'android.permission-group.HARDWARE_CONTROLS', 'description' : 'FM Radio', 'protectionLevel' : 'dangerous', 'label' : 'FM Radio'},
	'android.permission.READ_SOCIAL_STREAM' : {'permissionGroup' : 'android.permission-group.PERSONAL_INFO', 'description' : 'Allows the app to access and sync social updates from you and your friends. Be careful when sharing information -- this allows the app to read communications between you and your friends on social networks, regardless of confidentiality. Note: this permission may not be enforced on all social networks.', 'protectionLevel' : 'dangerous', 'label' : 'read your social stream'},
	'android.permission.DEVICE_POWER' : {'permissionGroup' : '', 'description' : 'Allows the app to turn the phone on or off.', 'protectionLevel' : 'signature', 'label' : 'power phone on or off'},
	'android.permission.ACCESS_MTP' : {'permissionGroup' : 'android.permission-group.HARDWARE_CONTROLS', 'description' : 'Allows access to the kernel MTP driver to implement the MTP USB protocol.', 'protectionLevel' : 'signature|system', 'label' : 'implement MTP protocol'},
	'android.permission.READ_NETWORK_USAGE_HISTORY' : {'permissionGroup' : '', 'description' : 'Allows the app to read historical network usage for specific networks and apps.', 'protectionLevel' : 'signature|system', 'label' : 'read historical network usage'},
	'android.permission.READ_SYNC_STATS' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows an app to read the sync stats for an account, including the history of sync events and how much data is synced.', 'protectionLevel' : 'normal', 'label' : 'read sync statistics'},
	'android.permission.SHUTDOWN' : {'permissionGroup' : '', 'description' : 'Puts the activity manager into a shutdown state. Does not perform a complete shutdown.', 'protectionLevel' : 'signature|system', 'label' : 'partial shutdown'},
	'android.permission.ACCESS_NETWORK_STATE' : {'permissionGroup' : 'android.permission-group.NETWORK', 'description' : 'Allows the app to view information about network connections such as which networks exist and are connected.', 'protectionLevel' : 'normal', 'label' : 'view network connections'},
	'android.permission.INTERNET' : {'permissionGroup' : 'android.permission-group.NETWORK', 'description' : 'Allows the app to create network sockets and use custom network protocols. The browser and other applications provide means to send data to the internet, so this permission is not required to send data to the internet.', 'protectionLevel' : 'dangerous', 'label' : 'full network access'},
	'android.permission.CHANGE_CONFIGURATION' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to change the current configuration, such as the locale or overall font size.', 'protectionLevel' : 'dangerous', 'label' : 'change system display settings'},
	'android.permission.READ_CONTACTS' : {'permissionGroup' : 'android.permission-group.PERSONAL_INFO', 'description' : 'Allows the app to read data about your contacts stored on your phone, including the frequency with which you\'ve called, emailed, or communicated in other ways with specific individuals. This permission allows apps to save your contact data, and malicious apps may share contact data without your knowledge.', 'protectionLevel' : 'dangerous', 'label' : 'read your contacts'},
	'android.permission.HARDWARE_TEST' : {'permissionGroup' : 'android.permission-group.HARDWARE_CONTROLS', 'description' : 'Allows the app to control various peripherals for the purpose of hardware testing.', 'protectionLevel' : 'signature', 'label' : 'test hardware'},
	'android.permission.SEND_DOWNLOAD_COMPLETED_INTENTS' : {'permissionGroup' : '', 'description' : 'Allows the app to send notifications about completed downloads. Malicious apps can use this to confuse other apps that download files.', 'protectionLevel' : 'signature', 'label' : 'Send download notifications.'},
	'com.android.launcher.permission.INSTALL_SHORTCUT' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows an app to add shortcuts without user intervention.', 'protectionLevel' : 'normal', 'label' : 'install shortcuts'},
	'android.permission.BIND_VPN_SERVICE' : {'permissionGroup' : '', 'description' : 'Allows the holder to bind to the top-level interface of a Vpn service. Should never be needed for normal apps.', 'protectionLevel' : 'signature', 'label' : 'bind to a VPN service'},
	'android.permission.WRITE_CALL_LOG' : {'permissionGroup' : 'android.permission-group.PERSONAL_INFO', 'description' : 'Allows the app to modify your phone\'s call log, including data about incoming and outgoing calls. Malicious apps may use this to erase or modify your call log.', 'protectionLevel' : 'dangerous', 'label' : 'write call log'},
	'android.permission.CHANGE_WIFI_MULTICAST_STATE' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to receive packets sent to all devices on a Wi-Fi network using multicast addresses, not just your phone. It uses more power than the non-multicast mode.', 'protectionLevel' : 'dangerous', 'label' : 'allow Wi-Fi Multicast reception'},
	'android.permission.STATUS_BAR' : {'permissionGroup' : '', 'description' : 'Allows the app to disable the status bar or add and remove system icons.', 'protectionLevel' : 'signature|system', 'label' : 'disable or modify status bar'},
	'android.permission.VIBRATE' : {'permissionGroup' : 'android.permission-group.HARDWARE_CONTROLS', 'description' : 'Allows the app to control the vibrator.', 'protectionLevel' : 'normal', 'label' : 'control vibration'},
	'android.permission.MODIFY_APPWIDGET_BIND_PERMISSIONS' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : '', 'protectionLevel' : 'signature|system', 'label' : ''},
	'android.permission.BIND_INPUT_METHOD' : {'permissionGroup' : '', 'description' : 'Allows the holder to bind to the top-level interface of an input method. Should never be needed for normal apps.', 'protectionLevel' : 'signature', 'label' : 'bind to an input method'},
	'android.permission.SET_TIME_ZONE' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to change the phone\'s time zone.', 'protectionLevel' : 'dangerous', 'label' : 'set time zone'},
	'android.permission.ACCESS_CACHE_FILESYSTEM' : {'permissionGroup' : '', 'description' : 'Allows the app to read and write the cache filesystem.', 'protectionLevel' : 'signature|system', 'label' : 'access the cache filesystem'},
	'android.permission.DOWNLOAD_CACHE_NON_PURGEABLE' : {'permissionGroup' : '', 'description' : 'Allows the app to download files to the download cache, which can\'t be automatically deleted when the download manager needs more space.', 'protectionLevel' : 'signatureOrSystem', 'label' : 'Reserve space in the download cache'},
	'android.permission.WRITE_SYNC_SETTINGS' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows an app to modify the sync settings for an account. For example, this can be used to enable sync of the People app with an account.', 'protectionLevel' : 'dangerous', 'label' : 'toggle sync on and off'},
	'android.permission.WRITE_USER_DICTIONARY' : {'permissionGroup' : 'android.permission-group.PERSONAL_INFO', 'description' : 'Allows the app to write new words into the user dictionary.', 'protectionLevel' : 'normal', 'label' : 'write to user-defined dictionary'},
	'android.permission.CRYPT_KEEPER' : {'permissionGroup' : '', 'description' : '', 'protectionLevel' : 'signature|system', 'label' : ''},
	'android.permission.READ_LOGS' : {'permissionGroup' : 'android.permission-group.DEVELOPMENT_TOOLS', 'description' : 'Allows the app to read from the system\'s various log files. This allows it to discover general information about what you are doing with the phone, potentially including personal or private information.', 'protectionLevel' : 'signature|system|development', 'label' : 'read sensitive log data'},
	'android.permission.DUMP' : {'permissionGroup' : 'android.permission-group.DEVELOPMENT_TOOLS', 'description' : 'Allows the app to retrieve internal state of the system. Malicious apps may retrieve a wide variety of private and secure information that they should never normally need.', 'protectionLevel' : 'signature|system|development', 'label' : 'retrieve system internal state'},
	'com.android.permission.WHITELIST_BLUETOOTH_DEVICE' : {'permissionGroup' : '', 'description' : 'Allows the app to temporarily whitelist a Bluetooth device, allowing that device to send files to this device without user confirmation.', 'protectionLevel' : 'signature', 'label' : 'Whitelist bluetooth device access.'},
	'android.permission.READ_FRAME_BUFFER' : {'permissionGroup' : '', 'description' : 'Allows the app to read the content of the frame buffer.', 'protectionLevel' : 'signature|system', 'label' : 'read frame buffer'},
	'android.permission.BIND_DEVICE_ADMIN' : {'permissionGroup' : '', 'description' : 'Allows the holder to send intents to a device administrator. Should never be needed for normal apps.', 'protectionLevel' : 'signature', 'label' : 'interact with a device admin'},
	'android.permission.FORCE_STOP_PACKAGES' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to forcibly stop other apps.', 'protectionLevel' : 'signature', 'label' : 'force stop other apps'},
	'com.android.frameworks.coretests.permission.TEST_DENIED' : {'permissionGroup' : '', 'description' : 'Used for running unit tests, for testing operations where we do not have the permission.', 'protectionLevel' : 'normal', 'label' : 'Test Denied'},
	'android.permission.WRITE_SECURE_SETTINGS' : {'permissionGroup' : 'android.permission-group.DEVELOPMENT_TOOLS', 'description' : 'Allows the app to modify the system\'s secure settings data. Not for use by normal apps.', 'protectionLevel' : 'signature|system|development', 'label' : 'modify secure system settings'},
	'android.permission.UPDATE_DEVICE_STATS' : {'permissionGroup' : '', 'description' : 'Allows the app to modify collected battery statistics. Not for use by normal apps.', 'protectionLevel' : 'signature|system', 'label' : 'modify battery statistics'},
	'android.permission.BROADCAST_PACKAGE_REMOVED' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to broadcast a notification that an app package has been removed. Malicious apps may use this to kill any other running app.', 'protectionLevel' : 'signature', 'label' : 'send package removed broadcast'},
	'android.permission.SYSTEM_ALERT_WINDOW' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to show system alert windows. Some alert windows may take over the entire screen.', 'protectionLevel' : 'dangerous', 'label' : 'draw over other apps'},
	'com.android.cts.permissionNotUsedWithSignature' : {'permissionGroup' : '', 'description' : '', 'protectionLevel' : 'signature', 'label' : ''},
	'android.permission.ACCESS_LOCATION_EXTRA_COMMANDS' : {'permissionGroup' : 'android.permission-group.LOCATION', 'description' : 'Allows the app to access extra location provider commands. This may allow the app to to interfere with the operation of the GPS or other location sources.', 'protectionLevel' : 'normal', 'label' : 'access extra location provider commands'},
	'android.permission.BRICK' : {'permissionGroup' : '', 'description' : 'Allows the app to disable the entire phone permanently. This is very dangerous.', 'protectionLevel' : 'signature', 'label' : 'permanently disable phone'},
	'android.permission.START_ANY_ACTIVITY' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to start any activity, regardless of permission protection or exported state.', 'protectionLevel' : 'signature', 'label' : 'start any activity'},
	'com.android.browser.permission.WRITE_HISTORY_BOOKMARKS' : {'permissionGroup' : 'android.permission-group.PERSONAL_INFO', 'description' : 'Allows the app to modify the Browser\'s history or bookmarks stored on your phone. This may allow the app to erase or modify Browser data. Note: this permission may note be enforced by third-party browsers or other applications with web browsing capabilities.', 'protectionLevel' : 'dangerous', 'label' : 'write web bookmarks and history'},
	'android.permission.CHANGE_WIFI_STATE' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to connect to and disconnect from Wi-Fi access points and to make changes to device configuration for Wi-Fi networks.', 'protectionLevel' : 'dangerous', 'label' : 'connect and disconnect from Wi-Fi'},
	'android.permission.RECORD_AUDIO' : {'permissionGroup' : 'android.permission-group.HARDWARE_CONTROLS', 'description' : 'Allows the app to record audio with the microphone. This permission allows the app to record audio at any time without your confirmation.', 'protectionLevel' : 'dangerous', 'label' : 'record audio'},
	'android.permission.MODIFY_PHONE_STATE' : {'permissionGroup' : 'android.permission-group.PHONE_CALLS', 'description' : 'Allows the app to control the phone features of the device. An app with this permission can switch networks, turn the phone radio on and off and the like without ever notifying you.', 'protectionLevel' : 'signature|system', 'label' : 'modify phone state'},
	'android.permission.READ_PROFILE' : {'permissionGroup' : 'android.permission-group.PERSONAL_INFO', 'description' : 'Allows the app to read personal profile information stored on your device, such as your name and contact information. This means the app can identify you and may send your profile information to others.', 'protectionLevel' : 'dangerous', 'label' : 'read your own contact card'},
	'android.permission.ACCOUNT_MANAGER' : {'permissionGroup' : 'android.permission-group.ACCOUNTS', 'description' : 'Allows the app to make calls to AccountAuthenticators.', 'protectionLevel' : 'signature', 'label' : 'act as the AccountManagerService'},
	'android.permission.SET_ANIMATION_SCALE' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to change the global animation speed (faster or slower animations) at any time.', 'protectionLevel' : 'signature|system|development', 'label' : 'modify global animation speed'},
	'android.permission.SET_PROCESS_LIMIT' : {'permissionGroup' : 'android.permission-group.DEVELOPMENT_TOOLS', 'description' : 'Allows the app to control the maximum number of processes that will run. Never needed for normal apps.', 'protectionLevel' : 'signature|system|development', 'label' : 'limit number of running processes'},
	'com.android.launcher.permission.PRELOAD_WORKSPACE' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : '', 'protectionLevel' : 'system|signature', 'label' : ''},
	'android.permission.WRITE_PROFILE' : {'permissionGroup' : 'android.permission-group.PERSONAL_INFO', 'description' : 'Allows the app to change or add to personal profile information stored on your device, such as your name and contact information. This means the app can identify you and may send your profile information to others.', 'protectionLevel' : 'dangerous', 'label' : 'modify your own contact card'},
	'android.permission.SET_DEBUG_APP' : {'permissionGroup' : 'android.permission-group.DEVELOPMENT_TOOLS', 'description' : 'Allows the app to turn on debugging for another app. Malicious apps may use this to kill other apps.', 'protectionLevel' : 'signature|system|development', 'label' : 'enable app debugging'},
	'android.permission.INSTALL_DRM' : {'permissionGroup' : '', 'description' : 'Allows app to install DRM-protected content.', 'protectionLevel' : 'normal', 'label' : 'Install DRM content.'},
	'android.permission.BLUETOOTH' : {'permissionGroup' : 'android.permission-group.NETWORK', 'description' : 'Allows the app to view the configuration of the Bluetooth on the phone, and to make and accept connections with paired devices.', 'protectionLevel' : 'dangerous', 'label' : 'pair with Bluetooth devices'},
	'android.permission.CAMERA' : {'permissionGroup' : 'android.permission-group.HARDWARE_CONTROLS', 'description' : 'Allows the app to take pictures and videos with the camera. This permission allows the app to use the camera at any time without your confirmation.', 'protectionLevel' : 'dangerous', 'label' : 'take pictures and videos'},
	'android.permission.SET_WALLPAPER_HINTS' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to set the system wallpaper size hints.', 'protectionLevel' : 'normal', 'label' : 'adjust your wallpaper size'},
	'android.permission.SET_ORIENTATION' : {'permissionGroup' : '', 'description' : 'Allows the app to change the rotation of the screen at any time. Should never be needed for normal apps.', 'protectionLevel' : 'signature', 'label' : 'change screen orientation'},
	'com.android.cts.permissionNormal' : {'permissionGroup' : '', 'description' : '', 'protectionLevel' : '', 'label' : ''},
	'android.permission.WAKE_LOCK' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to prevent the phone from going to sleep.', 'protectionLevel' : 'dangerous', 'label' : 'prevent phone from sleeping'},
	'com.android.frameworks.coretests.SIGNATURE' : {'permissionGroup' : 'android.permission-group.COST_MONEY', 'description' : '', 'protectionLevel' : 'signature', 'label' : ''},
	'android.permission.REBOOT' : {'permissionGroup' : '', 'description' : 'Allows the app to force the phone to reboot.', 'protectionLevel' : 'signature|system', 'label' : 'force phone reboot'},
	'android.permission.READ_PRIVILEGED_PHONE_STATE' : {'permissionGroup' : 'android.permission-group.PHONE_CALLS', 'description' : '', 'protectionLevel' : 'signature|system', 'label' : ''},
	'android.permission.BROADCAST_WAP_PUSH' : {'permissionGroup' : 'android.permission-group.MESSAGES', 'description' : 'Allows the app to broadcast a notification that a WAP PUSH message has been received. Malicious apps may use this to forge MMS message receipt or to silently replace the content of any webpage with malicious variants.', 'protectionLevel' : 'signature', 'label' : 'send WAP-PUSH-received broadcast'},
	'android.permission.SET_WALLPAPER_COMPONENT' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : '', 'protectionLevel' : 'signature|system', 'label' : ''},
	'android.permission.ACCESS_BLUETOOTH_SHARE' : {'permissionGroup' : '', 'description' : 'Allows the app to access the BluetoothShare manager and use it to transfer files.', 'protectionLevel' : 'signature', 'label' : 'Access download manager.'},
	'android.intent.category.MASTER_CLEAR.permission.C2D_MESSAGE' : {'permissionGroup' : '', 'description' : '', 'protectionLevel' : 'signature', 'label' : ''},
	'android.permission.UPDATE_LOCK' : {'permissionGroup' : '', 'description' : 'Allows the holder to offer information to the system about when would be a good time for a noninteractive reboot to upgrade the device.', 'protectionLevel' : 'signatureOrSystem', 'label' : 'discourage automatic device updates'},
	'android.permission.CHANGE_WIMAX_STATE' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to connect the phone to and disconnect the phone from WiMAX networks.', 'protectionLevel' : 'dangerous', 'label' : 'Change WiMAX state'},
	'com.android.browser.permission.READ_HISTORY_BOOKMARKS' : {'permissionGroup' : 'android.permission-group.PERSONAL_INFO', 'description' : 'Allows the app to read the history of all URLs that the Browser has visited, and all of the Browser\'s bookmarks. Note: this permission may not be enforced by third-party browsers or other applications with web browsing capabilities.', 'protectionLevel' : 'dangerous', 'label' : 'read your Web bookmarks and history'},
	'com.foo.mypermission' : {'permissionGroup' : '', 'description' : 'MyActivity', 'protectionLevel' : '', 'label' : 'MyActivity'},
	'android.permission.ACCESS_DRM' : {'permissionGroup' : '', 'description' : 'Allows app to access DRM-protected content.', 'protectionLevel' : 'signature', 'label' : 'Access DRM content.'},
	'android.permission.RECEIVE_SMS' : {'permissionGroup' : 'android.permission-group.MESSAGES', 'description' : 'Allows the app to receive and process SMS messages. This means the app could monitor or delete messages sent to your device without showing them to you.', 'protectionLevel' : 'dangerous', 'label' : 'receive text messages (SMS)'},
	'android.permission.WRITE_CONTACTS' : {'permissionGroup' : 'android.permission-group.PERSONAL_INFO', 'description' : 'Allows the app to modify the data about your contacts stored on your phone, including the frequency with which you\'ve called, emailed, or communicated in other ways with specific contacts. This permission allows apps to delete contact data.', 'protectionLevel' : 'dangerous', 'label' : 'modify your contacts'},
	'android.permission.CONTROL_LOCATION_UPDATES' : {'permissionGroup' : '', 'description' : 'Allows the app to enable/disable location update notifications from the radio. Not for use by normal apps.', 'protectionLevel' : 'signature|system', 'label' : 'control location update notifications'},
	'android.permission.BIND_APPWIDGET' : {'permissionGroup' : 'android.permission-group.PERSONAL_INFO', 'description' : 'Allows the app to tell the system which widgets can be used by which app. An app with this permission can give access to personal data to other apps. Not for use by normal apps.', 'protectionLevel' : 'signature|system', 'label' : 'choose widgets'},
	'com.android.frameworks.coretests.permission.TEST_GRANTED' : {'permissionGroup' : '', 'description' : 'Used for running unit tests, for testing operations where we have the permission.', 'protectionLevel' : 'normal', 'label' : 'Test Granted'},
	'android.permission.SIGNAL_PERSISTENT_PROCESSES' : {'permissionGroup' : 'android.permission-group.DEVELOPMENT_TOOLS', 'description' : 'Allows the app to request that the supplied signal be sent to all persistent processes.', 'protectionLevel' : 'signature|system|development', 'label' : 'send Linux signals to apps'},
	'android.permission.ASEC_CREATE' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to create internal storage.', 'protectionLevel' : 'signature', 'label' : 'create internal storage'},
	'android.permission.INSTALL_LOCATION_PROVIDER' : {'permissionGroup' : '', 'description' : 'Create mock location sources for testing or install a new location provider. This allows the app to override the location and/or status returned by other location sources such as GPS or location providers.', 'protectionLevel' : 'signature|system', 'label' : 'permission to install a location provider'},
	'android.permission.ACCESS_DOWNLOAD_MANAGER_ADVANCED' : {'permissionGroup' : '', 'description' : 'Allows the app to access the download manager\'s advanced functions. Malicious apps can use this to disrupt downloads and access private information.', 'protectionLevel' : 'signatureOrSystem', 'label' : 'Advanced download manager functions.'},
	'android.permission.WRITE_SETTINGS' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to modify the system\'s settings data. Malicious apps may corrupt your system\'s configuration.', 'protectionLevel' : 'dangerous', 'label' : 'modify system settings'},
	'android.permission.MASTER_CLEAR' : {'permissionGroup' : '', 'description' : 'Allows the app to completely reset the system to its factory settings, erasing all data, configuration, and installed apps.', 'protectionLevel' : 'signature|system', 'label' : 'reset system to factory defaults'},
	'android.permission.READ_INPUT_STATE' : {'permissionGroup' : '', 'description' : 'Allows the app to watch the keys you press even when interacting with another app (such as typing a password). Should never be needed for normal apps.', 'protectionLevel' : 'signature', 'label' : 'record what you type and actions you take'},
	'android.permission.INJECT_EVENTS' : {'permissionGroup' : '', 'description' : 'Allows the app to deliver its own input events (key presses, etc.) to other apps. Malicious apps may use this to take over the phone.', 'protectionLevel' : 'signature', 'label' : 'press keys and control buttons'},
	'android.permission.INTERNAL_SYSTEM_WINDOW' : {'permissionGroup' : '', 'description' : 'Allows the app to create windows that are intended to be used by the internal system user interface. Not for use by normal apps.', 'protectionLevel' : 'signature', 'label' : 'display unauthorized windows'},
	'android.permission.MANAGE_APP_TOKENS' : {'permissionGroup' : '', 'description' : 'Allows the app to create and manage their own tokens, bypassing their normal Z-ordering. Should never be needed for normal apps.', 'protectionLevel' : 'signature', 'label' : 'manage app tokens'},
	'com.android.email.permission.ACCESS_PROVIDER' : {'permissionGroup' : '', 'description' : 'Allows this application to access your email database, including received messages, sent messages, usernames, and passwords.', 'protectionLevel' : 'signature', 'label' : 'Access email provider data'},
	'android.permission.PACKAGE_VERIFICATION_AGENT' : {'permissionGroup' : '', 'description' : 'Allows the app to verify a package is installable.', 'protectionLevel' : 'signature|system', 'label' : 'verify packages'},
	'android.permission.CONFIRM_FULL_BACKUP' : {'permissionGroup' : '', 'description' : 'Allows the app to launch the full backup confirmation UI. Not to be used by any app.', 'protectionLevel' : 'signature', 'label' : 'confirm a full backup or restore operation'},
	'com.android.smspush.WAPPUSH_MANAGER_BIND' : {'permissionGroup' : '', 'description' : '', 'protectionLevel' : 'signatureOrSystem', 'label' : ''},
	'android.permission.ACCESS_WIMAX_STATE' : {'permissionGroup' : 'android.permission-group.NETWORK', 'description' : 'Allows the app to determine whether WiMAX is enabled and information about any WiMAX networks that are connected.', 'protectionLevel' : 'normal', 'label' : 'View WiMAX connections'},
	'com.android.launcher.permission.WRITE_SETTINGS' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to change the settings and shortcuts in Home.', 'protectionLevel' : 'normal', 'label' : 'write Home settings and shortcuts'},
	'android.permission.MODIFY_AUDIO_SETTINGS' : {'permissionGroup' : 'android.permission-group.HARDWARE_CONTROLS', 'description' : 'Allows the app to modify global audio settings such as volume and which speaker is used for output.', 'protectionLevel' : 'dangerous', 'label' : 'change your audio settings'},
	'android.permission.ASEC_ACCESS' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to get information on internal storage.', 'protectionLevel' : 'signature', 'label' : 'get information on internal storage'},
	'android.permission.USE_SIP' : {'permissionGroup' : 'android.permission-group.NETWORK', 'description' : 'Allows the app to use the SIP service to make/receive Internet calls.', 'protectionLevel' : 'dangerous', 'label' : 'make/receive Internet calls'},
	'android.permission.READ_CALL_LOG' : {'permissionGroup' : 'android.permission-group.PERSONAL_INFO', 'description' : 'Allows the app to read your phone\'s call log, including data about incoming and outgoing calls. This permission allows apps to save your call log data, and malicious apps may share call log data without your knowledge.', 'protectionLevel' : 'dangerous', 'label' : 'read call log'},
	'com.android.permission.HANDOVER_STATUS' : {'permissionGroup' : '', 'description' : 'Allows receiving handover transfer status information from Bluetooth.', 'protectionLevel' : 'signature', 'label' : 'Receive BT handover transfer broadcasts.'},
	'android.permission.WRITE_APN_SETTINGS' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to change network settings and to intercept and inspect all network traffic, for example to change the proxy and port of any APN. Malicious apps may monitor, redirect, or modify network packets without your knowledge.', 'protectionLevel' : 'signature|system', 'label' : 'change/intercept network settings and traffic'},
	'android.permission.ACCESS_SURFACE_FLINGER' : {'permissionGroup' : '', 'description' : 'Allows the app to use SurfaceFlinger low-level features.', 'protectionLevel' : 'signature', 'label' : 'access SurfaceFlinger'},
	'android.permission.MOVE_PACKAGE' : {'permissionGroup' : '', 'description' : 'Allows the app to move app resources from internal to external media and vice versa.', 'protectionLevel' : 'signature|system', 'label' : 'move app resources'},
	'android.permission.SERIAL_PORT' : {'permissionGroup' : '', 'description' : 'Allows the holder to access serial ports using the SerialManager API.', 'protectionLevel' : 'normal', 'label' : 'access serial ports'},
	'android.permission.NET_ADMIN' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : '', 'protectionLevel' : 'signature', 'label' : ''},
	'android.permission.ALLOW_ANY_CODEC_FOR_PLAYBACK' : {'permissionGroup' : '', 'description' : 'Allows the app to use any installed media decoder to decode for playback.', 'protectionLevel' : 'signature|system', 'label' : 'use any media decoder for playback'},
	'android.permission.MANAGE_USB' : {'permissionGroup' : 'android.permission-group.HARDWARE_CONTROLS', 'description' : 'Allows the app to manage preferences and permissions for USB devices.', 'protectionLevel' : 'signature|system', 'label' : 'manage preferences and permissions for USB devices'},
	'android.permission.CHANGE_BACKGROUND_DATA_SETTING' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to change the background data usage setting.', 'protectionLevel' : 'signature', 'label' : 'change background data usage setting'},
	'android.permission.PROCESS_OUTGOING_CALLS' : {'permissionGroup' : 'android.permission-group.PHONE_CALLS', 'description' : 'Allows the app to process outgoing calls and change the number to be dialed. This permission allows the app to monitor, redirect, or prevent outgoing calls.', 'protectionLevel' : 'dangerous', 'label' : 'reroute outgoing calls'},
	'android.permission.CALL_PRIVILEGED' : {'permissionGroup' : '', 'description' : 'Allows the app to call any phone number, including emergency numbers, without your intervention. Malicious apps may place unnecessary and illegal calls to emergency services.', 'protectionLevel' : 'signature|system', 'label' : 'directly call any phone numbers'},
	'android.permission.SET_PREFERRED_APPLICATIONS' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to modify your preferred apps. Malicious apps may silently change the apps that are run, spoofing your existing apps to collect private data from you.', 'protectionLevel' : 'signature', 'label' : 'set preferred apps'},
	'android.permission.WRITE_CALENDAR' : {'permissionGroup' : 'android.permission-group.PERSONAL_INFO', 'description' : 'Allows the app to add, remove, change events that you can modify on your phone, including those of friends or co-workers. This may allow the app to send messages that appear to come from calendar owners, or modify events without the owners\' knowledge.', 'protectionLevel' : 'dangerous', 'label' : 'add or modify calendar events and send email to guests without owners\' knowledge'},
	'android.permission.ACCESS_CONTENT_PROVIDERS_EXTERNALLY' : {'permissionGroup' : '', 'description' : 'Allows the holder to access content providers from the shell. Should never be needed for normal apps.', 'protectionLevel' : 'signature', 'label' : 'access content providers externally'},
	'android.permission.NFC' : {'permissionGroup' : 'android.permission-group.NETWORK', 'description' : 'Allows the app to communicate with Near Field Communication (NFC) tags, cards, and readers.', 'protectionLevel' : 'dangerous', 'label' : 'control Near Field Communication'},
	'android.permission.MANAGE_ACCOUNTS' : {'permissionGroup' : 'android.permission-group.ACCOUNTS', 'description' : 'Allows the app to perform operations like adding and removing accounts, and deleting their password.', 'protectionLevel' : 'dangerous', 'label' : 'add or remove accounts'},
	'android.permission.SEND_SMS' : {'permissionGroup' : 'android.permission-group.COST_MONEY', 'description' : 'Allows the app to send SMS messages. This may result in unexpected charges. Malicious apps may cost you money by sending messages without your confirmation.', 'protectionLevel' : 'dangerous', 'label' : 'send SMS messages'},
	'android.permission.BIND_REMOTEVIEWS' : {'permissionGroup' : '', 'description' : 'Allows the holder to bind to the top-level interface of a widget service. Should never be needed for normal apps.', 'protectionLevel' : 'signature|system', 'label' : 'bind to a widget service'},
	'android.permission.ACCESS_MOCK_LOCATION' : {'permissionGroup' : 'android.permission-group.LOCATION', 'description' : 'Create mock location sources for testing or install a new location provider. This allows the app to override the location and/or status returned by other location sources such as GPS or location providers.', 'protectionLevel' : 'dangerous', 'label' : 'mock location sources for testing'},
	'android.permission.BIND_ACCESSIBILITY_SERVICE' : {'permissionGroup' : '', 'description' : 'Allows the holder to bind to the top-level interface of an accessibility service. Should never be needed for normal apps.', 'protectionLevel' : 'signature', 'label' : 'bind to an accessibility service'},
	'android.permission.GET_DETAILED_TASKS' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to retrieve detailed information about currently and recently running tasks. Malicious apps may discover private information about other apps.', 'protectionLevel' : 'signature', 'label' : 'retrieve details of running apps'},
	'android.permission.WRITE_SMS' : {'permissionGroup' : 'android.permission-group.MESSAGES', 'description' : 'Allows the app to write to SMS messages stored on your phone or SIM card. Malicious apps may delete your messages.', 'protectionLevel' : 'dangerous', 'label' : 'edit your text messages (SMS or MMS)'},
	'android.permission.ACCESS_ALL_DOWNLOADS' : {'permissionGroup' : '', 'description' : 'Allows the app to view and modify all downloads initiated by any app on the system.', 'protectionLevel' : 'signature', 'label' : 'Access all system downloads'},
	'android.permission.DELETE_PACKAGES' : {'permissionGroup' : '', 'description' : 'Allows the app to delete Android packages. Malicious apps may use this to delete important apps.', 'protectionLevel' : 'signature|system', 'label' : 'delete apps'},
	'android.permission.COPY_PROTECTED_DATA' : {'permissionGroup' : '', 'description' : 'copy content', 'protectionLevel' : 'signature', 'label' : 'copy content'},
	'android.permission.ACCESS_CHECKIN_PROPERTIES' : {'permissionGroup' : '', 'description' : 'Allows the app read/write access to properties uploaded by the checkin service. Not for use by normal apps.', 'protectionLevel' : 'signature|system', 'label' : 'access checkin properties'},
	'android.permission.MOUNT_UNMOUNT_FILESYSTEMS' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to mount and unmount filesystems for removable storage.', 'protectionLevel' : 'dangerous', 'label' : 'access SD Card filesystem'},
	'android.permission.DOWNLOAD_WITHOUT_NOTIFICATION' : {'permissionGroup' : 'android.permission-group.NETWORK', 'description' : 'Allows the app to download files through the download manager without any notification being shown to the user.', 'protectionLevel' : 'normal', 'label' : 'download files without notification'},
	'android.permission.RETRIEVE_WINDOW_CONTENT' : {'permissionGroup' : 'android.permission-group.PERSONAL_INFO', 'description' : 'Allows the app to retrieve the content of the active window. Malicious apps may retrieve the entire window content and examine all its text except passwords.', 'protectionLevel' : 'signature|system', 'label' : 'retrieve screen content'},
	'com.android.email.permission.READ_ATTACHMENT' : {'permissionGroup' : 'android.permission-group.MESSAGES', 'description' : 'Allows this application to read your email attachments.', 'protectionLevel' : 'dangerous', 'label' : 'Read email attachments'},
	'android.permission.SET_TIME' : {'permissionGroup' : '', 'description' : 'Allows the app to change the phone\'s clock time.', 'protectionLevel' : 'signature|system', 'label' : 'set time'},
	'android.permission.BATTERY_STATS' : {'permissionGroup' : '', 'description' : 'Allows the app to modify collected battery statistics. Not for use by normal apps.', 'protectionLevel' : 'normal', 'label' : 'modify battery statistics'},
	'android.app.cts.permission.TEST_GRANTED' : {'permissionGroup' : '', 'description' : 'Used for running CTS tests, for testing operations where we have the permission.', 'protectionLevel' : 'normal', 'label' : 'Test Granted'},
	'android.permission.DIAGNOSTIC' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to read and write to any resource owned by the diag group; for example, files in /dev. This could potentially affect system stability and security. This should be ONLY be used for hardware-specific diagnostics by the manufacturer or operator.', 'protectionLevel' : 'signature', 'label' : 'read/write to resources owned by diag'},
	'com.android.cts.permissionAllowedWithSignature' : {'permissionGroup' : '', 'description' : '', 'protectionLevel' : 'signature', 'label' : ''},
	'android.permission.CALL_PHONE' : {'permissionGroup' : 'android.permission-group.COST_MONEY', 'description' : 'Allows the app to call phone numbers without your intervention. This may result in unexpected charges or calls. Note that this doesn\'t allow the app to call emergency numbers. Malicious apps may cost you money by making calls without your confirmation.', 'protectionLevel' : 'dangerous', 'label' : 'directly call phone numbers'},
	'android.permission.MOUNT_FORMAT_FILESYSTEMS' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to format removable storage.', 'protectionLevel' : 'dangerous', 'label' : 'erase SD Card'},
	'android.permission.READ_PHONE_STATE' : {'permissionGroup' : 'android.permission-group.PHONE_CALLS', 'description' : 'Allows the app to access the phone features of the device. This permission allows the app to determine the phone number and device IDs, whether a call is active, and the remote number connected by a call.', 'protectionLevel' : 'dangerous', 'label' : 'read phone status and identity'},
	'android.permission.GRANT_REVOKE_PERMISSIONS' : {'permissionGroup' : '', 'description' : 'Allows an application to grant or revoke specific permissions for it or other applications. Malicious applications may use this to access features you have not granted them.', 'protectionLevel' : 'signature', 'label' : 'grant or revoke permissions'},
	'android.permission.CLEAR_APP_USER_DATA' : {'permissionGroup' : '', 'description' : 'Allows the app to clear user data.', 'protectionLevel' : 'signature', 'label' : 'delete other apps\' data'},
	'android.permission.BROADCAST_SMS' : {'permissionGroup' : 'android.permission-group.MESSAGES', 'description' : 'Allows the app to broadcast a notification that an SMS message has been received. Malicious apps may use this to forge incoming SMS messages.', 'protectionLevel' : 'signature', 'label' : 'send SMS-received broadcast'},
	'android.permission.KILL_BACKGROUND_PROCESSES' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to end background processes of other apps. This may cause other apps to stop running.', 'protectionLevel' : 'normal', 'label' : 'close other apps'},
	'android.permission.STOP_APP_SWITCHES' : {'permissionGroup' : '', 'description' : 'Prevents the user from switching to another app.', 'protectionLevel' : 'signature|system', 'label' : 'prevent app switches'},
	'android.permission.ACCESS_WIFI_STATE' : {'permissionGroup' : 'android.permission-group.NETWORK', 'description' : 'Allows the app to view information about Wi-Fi networking, such as whether Wi-Fi is enabled and name of connected Wi-Fi devices.', 'protectionLevel' : 'normal', 'label' : 'view Wi-Fi connections'},
	'android.permission.RECEIVE_MMS' : {'permissionGroup' : 'android.permission-group.MESSAGES', 'description' : 'Allows the app to receive and process MMS messages. This means the app could monitor or delete messages sent to your device without showing them to you.', 'protectionLevel' : 'dangerous', 'label' : 'receive text messages (MMS)'},
	'android.permission.GLOBAL_SEARCH_CONTROL' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : '', 'protectionLevel' : 'signature', 'label' : ''},
	'android.permission.ACCESS_DOWNLOAD_MANAGER' : {'permissionGroup' : '', 'description' : 'Allows the app to access the download manager and to use it to download files. Malicious apps can use this to disrupt downloads and access private information.', 'protectionLevel' : 'signatureOrSystem', 'label' : 'Access download manager.'},
	'android.permission.STATUS_BAR_SERVICE' : {'permissionGroup' : '', 'description' : 'Allows the app to be the status bar.', 'protectionLevel' : 'signature', 'label' : 'status bar'},
	'android.permission.DELETE_CACHE_FILES' : {'permissionGroup' : '', 'description' : 'Allows the app to delete cache files.', 'protectionLevel' : 'signature|system', 'label' : 'delete other apps\' caches'},
	'android.permission.SET_POINTER_SPEED' : {'permissionGroup' : '', 'description' : 'Allows the app to change the mouse or trackpad pointer speed at any time. Should never be needed for normal apps.', 'protectionLevel' : 'signature', 'label' : 'change pointer speed'},
	'android.permission.RESTART_PACKAGES' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to end background processes of other apps. This may cause other apps to stop running.', 'protectionLevel' : 'normal', 'label' : 'close other apps'},
	'android.permission.MODIFY_NETWORK_ACCOUNTING' : {'permissionGroup' : '', 'description' : 'Allows the app to modify how network usage is accounted against apps. Not for use by normal apps.', 'protectionLevel' : 'signature|system', 'label' : 'modify network usage accounting'},
	'android.permission.GET_ACCOUNTS' : {'permissionGroup' : 'android.permission-group.ACCOUNTS', 'description' : 'Allows the app to get the list of accounts known by the phone. This may include any accounts created by applications you have installed.', 'protectionLevel' : 'normal', 'label' : 'find accounts on the device'},
	'android.permission.SUBSCRIBED_FEEDS_READ' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to get details about the currently synced feeds.', 'protectionLevel' : 'normal', 'label' : 'read subscribed feeds'},
	'android.permission.CHANGE_NETWORK_STATE' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to change the state of network connectivity.', 'protectionLevel' : 'dangerous', 'label' : 'change network connectivity'},
	'android.permission.READ_SYNC_SETTINGS' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to read the sync settings for an account. For example, this can determine whether the People app is synced with an account.', 'protectionLevel' : 'normal', 'label' : 'read sync settings'},
	'android.permission.DISABLE_KEYGUARD' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to disable the keylock and any associated password security. For example, the phone disables the keylock when receiving an incoming phone call, then re-enables the keylock when the call is finished.', 'protectionLevel' : 'dangerous', 'label' : 'disable your screen lock'},
	'android.permission.BIND_PACKAGE_VERIFIER' : {'permissionGroup' : '', 'description' : 'Allows the holder to make requests of package verifiers. Should never be needed for normal apps.', 'protectionLevel' : 'signature', 'label' : 'bind to a package verifier'},
	'com.android.launcher.permission.UNINSTALL_SHORTCUT' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to remove shortcuts without user intervention.', 'protectionLevel' : 'normal', 'label' : 'uninstall shortcuts'},
	'android.permission.USE_CREDENTIALS' : {'permissionGroup' : 'android.permission-group.ACCOUNTS', 'description' : 'Allows the app to request authentication tokens.', 'protectionLevel' : 'dangerous', 'label' : 'use accounts on the device'},
	'android.permission.SUBSCRIBED_FEEDS_WRITE' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to modify your currently synced feeds. Malicious apps may change your synced feeds.', 'protectionLevel' : 'dangerous', 'label' : 'write subscribed feeds'},
	'android.permission.READ_USER_DICTIONARY' : {'permissionGroup' : 'android.permission-group.PERSONAL_INFO', 'description' : 'Allows the app to read all words, names and phrases that the user may have stored in the user dictionary.', 'protectionLevel' : 'dangerous', 'label' : 'read terms you added to the dictionary'},
	'android.permission.WRITE_MEDIA_STORAGE' : {'permissionGroup' : 'android.permission-group.STORAGE', 'description' : 'Allows the app to modify the contents of the internal media storage.', 'protectionLevel' : 'signature|system', 'label' : 'modify/delete internal media storage contents'},
	'android.permission.FACTORY_TEST' : {'permissionGroup' : '', 'description' : 'Run as a low-level manufacturer test, allowing complete access to the phone hardware. Only available when a phone is running in manufacturer test mode.', 'protectionLevel' : 'signature', 'label' : 'run in factory test mode'},
	'android.permission.CHANGE_COMPONENT_ENABLED_STATE' : {'permissionGroup' : '', 'description' : 'Allows the app to change whether a component of another app is enabled or not. Malicious apps may use this to disable important phone capabilities. Care must be used with this permission, as it is possible to get app components into an unusable, inconsistent, or unstable state.', 'protectionLevel' : 'signature|system', 'label' : 'enable or disable app components'},
	'android.permission.RECEIVE_BOOT_COMPLETED' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to have itself started as soon as the system has finished booting. This can make it take longer to start the phone and allow the app to slow down the overall phone by always running.', 'protectionLevel' : 'normal', 'label' : 'run at startup'},
	'com.android.voicemail.permission.ADD_VOICEMAIL' : {'permissionGroup' : 'android.permission-group.PERSONAL_INFO', 'description' : 'Allows the app to add messages to your voicemail inbox.', 'protectionLevel' : 'dangerous', 'label' : 'add voicemail'},
	'android.permission.BACKUP' : {'permissionGroup' : '', 'description' : 'Allows the app to control the system\'s backup and restore mechanism. Not for use by normal apps.', 'protectionLevel' : 'signature|system', 'label' : 'control system backup and restore'},
	'com.android.voicemail.permission.READ_WRITE_ALL_VOICEMAIL' : {'permissionGroup' : 'android.permission-group.PERSONAL_INFO', 'description' : 'Allows the app to store and retrieve all voicemails that this device can access.', 'protectionLevel' : 'signature', 'label' : 'Access all voicemails'},
	'android.permission.BLUETOOTH_ADMIN' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to configure the local Bluetooth phone, and to discover and pair with remote devices.', 'protectionLevel' : 'dangerous', 'label' : 'access Bluetooth settings'},
	'android.permission.ACCESS_FINE_LOCATION' : {'permissionGroup' : 'android.permission-group.LOCATION', 'description' : 'Access precise location sources such as the Global Positioning System on the phone. When location services are available and turned on, this permission allows the app to determine your precise location.', 'protectionLevel' : 'dangerous', 'label' : 'precise (GPS) location'},
	'android.permission.ASEC_RENAME' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to rename internal storage.', 'protectionLevel' : 'signature', 'label' : 'rename internal storage'},
	'android.permission.PERSISTENT_ACTIVITY' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to make parts of itself persistent in memory. This can limit memory available to other apps slowing down the phone.', 'protectionLevel' : 'dangerous', 'label' : 'make app always run'},
	'android.permission.REORDER_TASKS' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to move tasks to the foreground and background. The app may do this without your input.', 'protectionLevel' : 'dangerous', 'label' : 'reorder running apps'},
	'android.permission.BIND_TEXT_SERVICE' : {'permissionGroup' : '', 'description' : 'Allows the holder to bind to the top-level interface of a text service(e.g. SpellCheckerService). Should never be needed for normal apps.', 'protectionLevel' : 'signature', 'label' : 'bind to a text service'},
	'android.permission.RECEIVE_WAP_PUSH' : {'permissionGroup' : 'android.permission-group.MESSAGES', 'description' : 'Allows the app to receive and process WAP messages. This permission includes the ability to monitor or delete messages sent to you without showing them to you.', 'protectionLevel' : 'dangerous', 'label' : 'receive text messages (WAP)'},
	'com.foo.mypermission2' : {'permissionGroup' : '', 'description' : 'MyActivity', 'protectionLevel' : '', 'label' : 'MyActivity'},
	'android.permission.EXPAND_STATUS_BAR' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to expand or collapse the status bar.', 'protectionLevel' : 'normal', 'label' : 'expand/collapse status bar'},
	'android.permission.SET_WALLPAPER' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to set the system wallpaper.', 'protectionLevel' : 'normal', 'label' : 'set wallpaper'},
	'android.permission.ASEC_DESTROY' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to destroy internal storage.', 'protectionLevel' : 'signature', 'label' : 'destroy internal storage'},
	'android.permission.CONNECTIVITY_INTERNAL' : {'permissionGroup' : 'android.permission-group.NETWORK', 'description' : '', 'protectionLevel' : 'signature|system', 'label' : ''},
	'android.permission.SET_SCREEN_COMPATIBILITY' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to control the screen compatibility mode of other applications. Malicious applications may break the behavior of other applications.', 'protectionLevel' : 'signature', 'label' : 'set screen compatibility'},
	'android.permission.WRITE_EXTERNAL_STORAGE' : {'permissionGroup' : 'android.permission-group.STORAGE', 'description' : 'Allows the app to write to the SD card.', 'protectionLevel' : 'dangerous', 'label' : 'modify or delete the contents of your SD card'},
	'android.permission.GET_PACKAGE_SIZE' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to retrieve its code, data, and cache sizes', 'protectionLevel' : 'normal', 'label' : 'measure app storage space'},
	'com.android.frameworks.coretests.DANGEROUS' : {'permissionGroup' : 'android.permission-group.COST_MONEY', 'description' : '', 'protectionLevel' : 'dangerous', 'label' : ''},
	'android.permission.WRITE_SOCIAL_STREAM' : {'permissionGroup' : 'android.permission-group.PERSONAL_INFO', 'description' : 'Allows the app to display social updates from your friends. Be careful when sharing information -- this allows the app to produce messages that may appear to come from a friend. Note: this permission may not be enforced on all social networks.', 'protectionLevel' : 'dangerous', 'label' : 'write to your social stream'},
	'android.permission.READ_EXTERNAL_STORAGE' : {'permissionGroup' : 'android.permission-group.DEVELOPMENT_TOOLS', 'description' : 'Allows the app to test a permission for the SD card that will be available on future devices.', 'protectionLevel' : 'normal', 'label' : 'test access to protected storage'},
	'android.permission.WRITE_GSERVICES' : {'permissionGroup' : '', 'description' : 'Allows the app to modify the Google services map. Not for use by normal apps.', 'protectionLevel' : 'signature|system', 'label' : 'modify the Google services map'},
	'android.permission.ASEC_MOUNT_UNMOUNT' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to mount/unmount internal storage.', 'protectionLevel' : 'signature', 'label' : 'mount/unmount internal storage'},
	'android.permission.INSTALL_PACKAGES' : {'permissionGroup' : '', 'description' : 'Allows the app to install new or updated Android packages. Malicious apps may use this to add new apps with arbitrarily powerful permissions.', 'protectionLevel' : 'signature|system', 'label' : 'directly install apps'},
	'android.permission.SET_KEYBOARD_LAYOUT' : {'permissionGroup' : '', 'description' : 'Allows the app to change the keyboard layout. Should never be needed for normal apps.', 'protectionLevel' : 'signature', 'label' : 'change keyboard layout'},
	'android.permission.AUTHENTICATE_ACCOUNTS' : {'permissionGroup' : 'android.permission-group.ACCOUNTS', 'description' : 'Allows the app to use the account authenticator capabilities of the AccountManager, including creating accounts and getting and setting their passwords.', 'protectionLevel' : 'dangerous', 'label' : 'create accounts and set passwords'},
	'android.permission.RECEIVE_EMERGENCY_BROADCAST' : {'permissionGroup' : 'android.permission-group.MESSAGES', 'description' : 'Allows the app to receive and process emergency broadcast messages. This permission is only available to system apps.', 'protectionLevel' : 'signature|system', 'label' : 'receive emergency broadcasts'},
	'com.android.launcher.permission.READ_SETTINGS' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to read the settings and shortcuts in Home.', 'protectionLevel' : 'normal', 'label' : 'read Home settings and shortcuts'},
	'com.android.alarm.permission.SET_ALARM' : {'permissionGroup' : 'android.permission-group.PERSONAL_INFO', 'description' : 'Allows the app to set an alarm in an installed alarm clock app. Some alarm clock apps may not implement this feature.', 'protectionLevel' : 'normal', 'label' : 'set an alarm'},
	'ti.permission.FMRX_ADMIN' : {'permissionGroup' : 'android.permission-group.HARDWARE_CONTROLS', 'description' : 'FM Radio', 'protectionLevel' : 'dangerous', 'label' : 'FM Radio'},
	'android.permission.PERFORM_CDMA_PROVISIONING' : {'permissionGroup' : '', 'description' : 'Allows the app to start CDMA provisioning. Malicious apps may unnecessarily start CDMA provisioning.', 'protectionLevel' : 'signature|system', 'label' : 'directly start CDMA phone setup'},
	'com.android.browser.permission.PRELOAD' : {'permissionGroup' : '', 'description' : '', 'protectionLevel' : 'signatureOrSystem', 'label' : 'Preload results'},
	'android.permission.GET_TASKS' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to retrieve information about currently and recently running tasks. This may allow the app to discover information about which applications are used on the device.', 'protectionLevel' : 'dangerous', 'label' : 'retrieve running apps'},
	'android.permission.READ_CELL_BROADCASTS' : {'permissionGroup' : 'android.permission-group.MESSAGES', 'description' : 'Allows the app to read cell broadcast messages received by your device. Cell broadcast alerts are delivered in some locations to warn you of emergency situations. Malicious apps may interfere with the performance or operation of your device when an emergency cell broadcast is received.', 'protectionLevel' : 'dangerous', 'label' : 'read cell broadcast messages'},
	'android.permission.SET_ACTIVITY_WATCHER' : {'permissionGroup' : '', 'description' : 'Allows the app to monitor and control how the system launches activities. Malicious apps may completely compromise the system. This permission is only needed for development, never for normal use.', 'protectionLevel' : 'signature', 'label' : 'monitor and control all app launching'},
	'com.android.frameworks.coretests.NORMAL' : {'permissionGroup' : 'android.permission-group.COST_MONEY', 'description' : '', 'protectionLevel' : 'normal', 'label' : ''},
	'android.permission.READ_SMS' : {'permissionGroup' : 'android.permission-group.MESSAGES', 'description' : 'Allows the app to read SMS messages stored on your phone or SIM card. This allows the app to read all SMS messages, regardless of content or confidentiality.', 'protectionLevel' : 'dangerous', 'label' : 'read your text messages (SMS or MMS)'},
	'android.permission.BROADCAST_STICKY' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to send sticky broadcasts, which remain after the broadcast ends. Excessive use may make the phone slow or unstable by causing it to use too much memory.', 'protectionLevel' : 'normal', 'label' : 'send sticky broadcast'},
	'android.permission.GLOBAL_SEARCH' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : '', 'protectionLevel' : 'signature|system', 'label' : ''},
	'android.permission.SEND_SMS_NO_CONFIRMATION' : {'permissionGroup' : 'android.permission-group.COST_MONEY', 'description' : 'Allows the app to send SMS messages. This may result in unexpected charges. Malicious apps may cost you money by sending messages without your confirmation.', 'protectionLevel' : 'signature|system', 'label' : 'send SMS messages with no confirmation'},
	'com.android.cts.permissionWithSignature' : {'permissionGroup' : '', 'description' : '', 'protectionLevel' : 'signature', 'label' : ''},
	'android.permission.REMOVE_TASKS' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to remove tasks and kill their apps. Malicious apps may disrupt the behavior of other apps.', 'protectionLevel' : 'signature', 'label' : 'stop running apps'},
	'android.permission.PACKAGE_USAGE_STATS' : {'permissionGroup' : '', 'description' : 'Allows the app to modify collected component usage statistics. Not for use by normal apps.', 'protectionLevel' : 'signature|system', 'label' : 'update component usage statistics'},
	'android.permission.SET_ALWAYS_FINISH' : {'permissionGroup' : 'android.permission-group.DEVELOPMENT_TOOLS', 'description' : 'Allows the app to control whether activities are always finished as soon as they go to the background. Never needed for normal apps.', 'protectionLevel' : 'signature|system|development', 'label' : 'force background apps to close'},
	'android.permission.CLEAR_APP_CACHE' : {'permissionGroup' : 'android.permission-group.SYSTEM_TOOLS', 'description' : 'Allows the app to free phone storage by deleting files in app cache directory. Access is very restricted usually to system process.', 'protectionLevel' : 'dangerous', 'label' : 'delete all app cache data'},
	'android.permission.MANAGE_NETWORK_POLICY' : {'permissionGroup' : '', 'description' : 'Allows the app to manage network policies and define app-specific rules.', 'protectionLevel' : 'signature', 'label' : 'manage network policy'},
	'android.permission.FLASHLIGHT' : {'permissionGroup' : 'android.permission-group.HARDWARE_CONTROLS', 'description' : 'Allows the app to control the flashlight.', 'protectionLevel' : 'normal', 'label' : 'control flashlight'},
}

AOSP_PERMISSION_GROUPS = {
	'android.permission-group.NETWORK' : {'description' : 'Access various network features.', 'label' : 'Network communication'},
	'android.permission-group.STORAGE' : {'description' : 'Access the SD card.', 'label' : 'Storage'},
	'android.permission-group.MESSAGES' : {'description' : 'Read and write your SMS, email, and other messages.', 'label' : 'Your messages'},
	'android.permission-group.PERSONAL_INFO' : {'description' : 'Direct access to your contacts and calendar stored on the phone.', 'label' : 'Your personal information'},
	'android.permission-group.DEVELOPMENT_TOOLS' : {'description' : 'Features only needed for app developers.', 'label' : 'Development tools'},
	'android.permission-group.COST_MONEY' : {'description' : '', 'label' : ''},
	'android.permission-group.ACCOUNTS' : {'description' : 'Access the available accounts.', 'label' : 'Your accounts'},
	'android.permission-group.LOCATION' : {'description' : 'Monitor your physical location.', 'label' : 'Your location'},
	'android.permission-group.HARDWARE_CONTROLS' : {'description' : 'Direct access to hardware on the handset.', 'label' : 'Hardware controls'},
	'android.permission-group.SYSTEM_TOOLS' : {'description' : 'Lower-level access and control of the system.', 'label' : 'System tools'},
	'android.permission-group.PHONE_CALLS' : {'description' : 'Monitor, record, and process phone calls.', 'label' : 'Phone calls'},
}
#################################################





# This file is part of Androguard.
#
# Copyright (C) 2012, Anthony Desnos <desnos at t0t0.fr>
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os
import logging
import types
import random
import string
import imp

ANDROGUARD_VERSION = "3.0"

from androguard.core.api_specific_resources.aosp_permissions.aosp_permissions import AOSP_PERMISSIONS
from androguard.core.api_specific_resources.api_permission_mappings.api_permission_mappings import AOSP_PERMISSIONS_MAPPINGS

def is_ascii_problem(s):
    try:
        s.decode("ascii")
        return False
    except UnicodeDecodeError:
        return True


class Color(object):
    Normal = "\033[0m"
    Black = "\033[30m"
    Red = "\033[31m"
    Green = "\033[32m"
    Yellow = "\033[33m"
    Blue = "\033[34m"
    Purple = "\033[35m"
    Cyan = "\033[36m"
    Grey = "\033[37m"
    Bold = "\033[1m"

CONF = {
    "BIN_DED": "ded.sh",
    "PATH_DED": "./decompiler/ded/",
    "PATH_DEX2JAR": "./decompiler/dex2jar/",
    "BIN_DEX2JAR": "dex2jar.sh",
    "PATH_JAD": "./decompiler/jad/",
    "BIN_JAD": "jad",
    "BIN_WINEJAD": "jad.exe",
    "PATH_FERNFLOWER": "./decompiler/fernflower/",
    "BIN_FERNFLOWER": "fernflower.jar",
    "OPTIONS_FERNFLOWER": {"dgs": '1', "asc": '1'},
    "PRETTY_SHOW": 1,

    "TMP_DIRECTORY": "/tmp/",

    # Full python or mix python/c++ (native)
    #"ENGINE" : "automatic",
    "ENGINE": "python",

    "RECODE_ASCII_STRING": False,
    "RECODE_ASCII_STRING_METH": None,

    "DEOBFUSCATED_STRING": True,
#    "DEOBFUSCATED_STRING_METH" : get_deobfuscated_string,

    "PATH_JARSIGNER": "jarsigner",

    "COLORS": {
        "OFFSET": Color.Yellow,
        "OFFSET_ADDR": Color.Green,
        "INSTRUCTION_NAME": Color.Yellow,
        "BRANCH_FALSE": Color.Red,
        "BRANCH_TRUE": Color.Green,
        "BRANCH": Color.Blue,
        "EXCEPTION": Color.Cyan,
        "BB": Color.Purple,
        "NOTE": Color.Red,
        "NORMAL": Color.Normal,

        "OUTPUT": {
            "normal": Color.Normal,
            "registers": Color.Normal,
            "literal": Color.Green,
            "offset": Color.Purple,
            "raw": Color.Red,
            "string": Color.Red,
            "meth": Color.Cyan,
            "type": Color.Blue,
            "field": Color.Green,
        }
    },

    "PRINT_FCT": sys.stdout.write,
    "LAZY_ANALYSIS": False,
    "MAGIC_PATH_FILE": None,

    "DEFAULT_API" : 19,
}


def default_colors(obj):
  CONF["COLORS"]["OFFSET"] = obj.Yellow
  CONF["COLORS"]["OFFSET_ADDR"] = obj.Green
  CONF["COLORS"]["INSTRUCTION_NAME"] = obj.Yellow
  CONF["COLORS"]["BRANCH_FALSE"] = obj.Red
  CONF["COLORS"]["BRANCH_TRUE"] = obj.Green
  CONF["COLORS"]["BRANCH"] = obj.Blue
  CONF["COLORS"]["EXCEPTION"] = obj.Cyan
  CONF["COLORS"]["BB"] = obj.Purple
  CONF["COLORS"]["NOTE"] = obj.Red
  CONF["COLORS"]["NORMAL"] = obj.Normal

  CONF["COLORS"]["OUTPUT"]["normal"] = obj.Normal
  CONF["COLORS"]["OUTPUT"]["registers"] = obj.Normal
  CONF["COLORS"]["OUTPUT"]["literal"] = obj.Green
  CONF["COLORS"]["OUTPUT"]["offset"] = obj.Purple
  CONF["COLORS"]["OUTPUT"]["raw"] = obj.Red
  CONF["COLORS"]["OUTPUT"]["string"] = obj.Red
  CONF["COLORS"]["OUTPUT"]["meth"] = obj.Cyan
  CONF["COLORS"]["OUTPUT"]["type"] = obj.Blue
  CONF["COLORS"]["OUTPUT"]["field"] = obj.Green


def disable_colors():
  """ Disable colors from the output (color = normal)"""
  for i in CONF["COLORS"]:
    if isinstance(CONF["COLORS"][i], dict):
        for j in CONF["COLORS"][i]:
            CONF["COLORS"][i][j] = Color.normal
    else:
        CONF["COLORS"][i] = Color.normal


def remove_colors():
  """ Remove colors from the output (no escape sequences)"""
  for i in CONF["COLORS"]:
    if isinstance(CONF["COLORS"][i], dict):
        for j in CONF["COLORS"][i]:
            CONF["COLORS"][i][j] = ""
    else:
        CONF["COLORS"][i] = ""


def enable_colors(colors):
  for i in colors:
    CONF["COLORS"][i] = colors[i]


def save_colors():
  c = {}
  for i in CONF["COLORS"]:
    if isinstance(CONF["COLORS"][i], dict):
        c[i] = {}
        for j in CONF["COLORS"][i]:
            c[i][j] = CONF["COLORS"][i][j]
    else:
        c[i] = CONF["COLORS"][i]
  return c


def long2int(l):
    if l > 0x7fffffff:
        l = (0x7fffffff & l) - 0x80000000
    return l


def long2str(l):
    """Convert an integer to a string."""
    if type(l) not in (types.IntType, types.LongType):
        raise ValueError, 'the input must be an integer'

    if l < 0:
        raise ValueError, 'the input must be greater than 0'
    s = ''
    while l:
        s = s + chr(l & 255L)
        l >>= 8

    return s

def str2long(s):
    """Convert a string to a long integer."""
    if type(s) not in (types.StringType, types.UnicodeType):
        raise ValueError, 'the input must be a string'

    l = 0L
    for i in s:
        l <<= 8
        l |= ord(i)

    return l

def random_string():
    return random.choice( string.letters ) + ''.join([ random.choice(string.letters + string.digits) for i in range(10 - 1) ] )

def is_android(filename):
    """Return the type of the file

        @param filename : the filename
        @rtype : "APK", "DEX", "ELF", None
    """
    if not filename:
        return None

    val = None
    with open(filename, "r") as fd:
        f_bytes = fd.read()
        val = is_android_raw( f_bytes )

    return val

def is_android_raw(raw):
    val = None

    if raw[0:2] == "PK" or ('AndroidManifest.xml' in raw and 'META-INF/MANIFEST.MF' in raw):
        val = "APK"
    elif raw[0:3] == "dex":
        val = "DEX"
    elif raw[0:3] == "dey":
        val = "DEY"
    elif raw[0:7] == "\x7fELF\x01\x01\x01":
        val = "ELF"
    elif raw[0:4] == "\x03\x00\x08\x00":
        val = "AXML"
    elif raw[0:4] == "\x02\x00\x0C\x00":
        val = "ARSC"

    return val

def is_valid_android_raw(raw):
  return raw.find("classes.dex") != -1

# from scapy
log_andro = logging.getLogger("andro")
console_handler = logging.StreamHandler()
console_handler.setFormatter(logging.Formatter("%(levelname)s: %(message)s"))
log_andro.addHandler(console_handler)
log_runtime = logging.getLogger("andro.runtime")          # logs at runtime
log_interactive = logging.getLogger("andro.interactive")  # logs in interactive functions
log_loading = logging.getLogger("andro.loading")          # logs when loading andro

def set_lazy():
  CONF["LAZY_ANALYSIS"] = True

def set_debug():
    log_andro.setLevel( logging.DEBUG )

def set_info():
    log_andro.setLevel(logging.INFO)

def get_debug():
    return log_andro.getEffectiveLevel() == logging.DEBUG

def warning(x):
    log_runtime.warning(x)
    import traceback
    traceback.print_exc()

def error(x):
    log_runtime.error(x)
    raise()

def debug(x):
    log_runtime.debug(x)

def info(x):
    log_runtime.info(x)

def set_options(key, value):
    CONF[ key ] = value

def save_to_disk(buff, output):
    with open(output, "w") as fd:
        fd.write(buff)

def rrmdir( directory ):
    for root, dirs, files in os.walk(directory, topdown=False):
        for name in files:
            os.remove(os.path.join(root, name))
        for name in dirs:
            os.rmdir(os.path.join(root, name))
    os.rmdir( directory )


def make_color_tuple( color ):
    """
    turn something like "#000000" into 0,0,0
    or "#FFFFFF into "255,255,255"
    """
    R = color[1:3]
    G = color[3:5]
    B = color[5:7]

    R = int(R, 16)
    G = int(G, 16)
    B = int(B, 16)

    return R,G,B

def interpolate_tuple( startcolor, goalcolor, steps ):
    """
    Take two RGB color sets and mix them over a specified number of steps.  Return the list
    """
    # white

    R = startcolor[0]
    G = startcolor[1]
    B = startcolor[2]

    targetR = goalcolor[0]
    targetG = goalcolor[1]
    targetB = goalcolor[2]

    DiffR = targetR - R
    DiffG = targetG - G
    DiffB = targetB - B

    buffer = []

    for i in range(0, steps +1):
        iR = R + (DiffR * i / steps)
        iG = G + (DiffG * i / steps)
        iB = B + (DiffB * i / steps)

        hR = string.replace(hex(iR), "0x", "")
        hG = string.replace(hex(iG), "0x", "")
        hB = string.replace(hex(iB), "0x", "")

        if len(hR) == 1:
            hR = "0" + hR
        if len(hB) == 1:
            hB = "0" + hB

        if len(hG) == 1:
            hG = "0" + hG

        color = string.upper("#"+hR+hG+hB)
        buffer.append(color)

    return buffer

def color_range( startcolor, goalcolor, steps ):
    """
    wrapper for interpolate_tuple that accepts colors as html ("#CCCCC" and such)
    """
    start_tuple = make_color_tuple(startcolor)
    goal_tuple = make_color_tuple(goalcolor)

    return interpolate_tuple(start_tuple, goal_tuple, steps)


def load_api_specific_resource_module(resource_name, api):
    if resource_name == "aosp_permissions":
        module = AOSP_PERMISSIONS
    elif resource_name == "api_permission_mappings":
        module = AOSP_PERMISSIONS_MAPPINGS
    else:
        error("Invalid resource: %s" % resource_name)

    if not api:
        api = CONF["DEFAULT_API"]
    value = module.get(api)
    if value:
        return value
    return module.get('9')



# This file is part of Androguard.
#
# Copyright (C) 2012, Anthony Desnos <desnos at t0t0.fr>
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re, random, cPickle, collections

from androguard.core.androconf import error, warning, debug, is_ascii_problem,\
    load_api_specific_resource_module
from androguard.core.bytecodes import dvm
from androguard.core.bytecodes.api_permissions import DVM_PERMISSIONS_BY_PERMISSION, DVM_PERMISSIONS_BY_ELEMENT

class ContextField(object):
    def __init__(self, mode):
        self.mode = mode
        self.details = []

    def set_details(self, details):
        for i in details:
            self.details.append( i )

class ContextMethod(object):
    def __init__(self):
        self.details = []

    def set_details(self, details):
        for i in details:
            self.details.append( i )

class ExternalFM(object):
    def __init__(self, class_name, name, descriptor):
        self.class_name = class_name
        self.name = name
        self.descriptor = descriptor

    def get_class_name(self):
        return self.class_name

    def get_name(self):
        return self.name

    def get_descriptor(self):
        return self.descriptor

class ToString(object):
    def __init__(self, tab):
        self.__tab = tab
        self.__re_tab = {}

        for i in self.__tab:
            self.__re_tab[i] = []
            for j in self.__tab[i]:
                self.__re_tab[i].append( re.compile( j ) )

        self.__string = ""

    def push(self, name):
        for i in self.__tab:
            for j in self.__re_tab[i]:
                if j.match(name) != None:
                    if len(self.__string) > 0:
                        if i == 'O' and self.__string[-1] == 'O':
                            continue
                    self.__string += i

    def get_string(self):
        return self.__string

class BreakBlock(object):
    def __init__(self, _vm, idx):
        self._vm = _vm
        self._start = idx
        self._end = self._start

        self._ins = []

        self._ops = []

        self._fields = {}
        self._methods = {}


    def get_ops(self):
        return self._ops

    def get_fields(self):
        return self._fields

    def get_methods(self):
        return self._methods

    def push(self, ins):
        self._ins.append(ins)
        self._end += ins.get_length()

    def get_start(self):
        return self._start

    def get_end(self):
        return self._end

    def show(self):
        for i in self._ins:
            print "\t\t",
            i.show(0)

DVM_FIELDS_ACCESS = {
      "iget" : "R",
      "iget-wide" : "R",
      "iget-object" : "R",
      "iget-boolean" : "R",
      "iget-byte" : "R",
      "iget-char" : "R",
      "iget-short" : "R",

      "iput" : "W",
      "iput-wide" : "W",
      "iput-object" : "W",
      "iput-boolean" : "W",
      "iput-byte" : "W",
      "iput-char" : "W",
      "iput-short" : "W",

      "sget" : "R",
      "sget-wide" : "R",
      "sget-object" : "R",
      "sget-boolean" : "R",
      "sget-byte" : "R",
      "sget-char" : "R",
      "sget-short" : "R",

      "sput" : "W",
      "sput-wide" : "W",
      "sput-object" : "W",
      "sput-boolean" : "W",
      "sput-byte" : "W",
      "sput-char" : "W",
      "sput-short" : "W",
   }


class DVMBasicBlock(object):
    """
        A simple basic block of a dalvik method
    """
    def __init__(self, start, vm, method, context):
        self.__vm = vm
        self.method = method
        self.context = context

        self.last_length = 0
        self.nb_instructions = 0

        self.fathers = []
        self.childs = []

        self.start = start
        self.end = self.start

        self.special_ins = {}

        self.name = "%s-BB@0x%x" % (self.method.get_name(), self.start)
        self.exception_analysis = None

        self.tainted_variables = self.context.get_tainted_variables()
        self.tainted_packages = self.context.get_tainted_packages()

        self.notes = []

    def get_notes(self):
        return self.notes

    def set_notes(self, value):
        self.notes = [value]

    def add_note(self, note):
        self.notes.append(note)

    def clear_notes(self):
        self.notes = []

    def get_instructions(self):
      """
        Get all instructions from a basic block.

        :rtype: Return all instructions in the current basic block
      """
      tmp_ins = []
      idx = 0
      for i in self.method.get_instructions():
        if idx >= self.start and idx < self.end:
          tmp_ins.append(i)

        idx += i.get_length()
      return tmp_ins

    def get_nb_instructions(self):
        return self.nb_instructions

    def get_method(self):
        return self.method

    def get_name(self):
        return "%s-BB@0x%x" % (self.method.get_name(), self.start)

    def get_start(self):
        return self.start

    def get_end(self):
        return self.end

    def get_last(self):
        return self.get_instructions()[-1]

    def get_next(self):
        """
            Get next basic blocks

            :rtype: a list of the next basic blocks
        """
        return self.childs

    def get_prev(self):
        """
            Get previous basic blocks

            :rtype: a list of the previous basic blocks
        """
        return self.fathers

    def set_fathers(self, f):
        self.fathers.append(f)

    def get_last_length(self):
      return self.last_length

    def set_childs(self, values):
        #print self, self.start, self.end, values
        if values == []:
            next_block = self.context.get_basic_block( self.end + 1 )
            if next_block != None:
                self.childs.append( ( self.end - self.get_last_length(), self.end, next_block ) )
        else:
            for i in values:
                if i != -1:
                    next_block = self.context.get_basic_block( i )
                    if next_block != None:
                        self.childs.append( ( self.end - self.get_last_length(), i, next_block) )

        for c in self.childs:
            if c[2] != None:
                c[2].set_fathers( ( c[1], c[0], self ) )

    def push(self, i):
      try:
            self.nb_instructions += 1
            idx = self.end
            self.last_length = i.get_length()
            self.end += self.last_length

            op_value = i.get_op_value()

            # field access
            if (op_value >= 0x52 and op_value <= 0x6d):
                desc = self.__vm.get_cm_field(i.get_ref_kind())
                if self.tainted_variables != None:
                    self.tainted_variables.push_info(TAINTED_FIELD, desc, DVM_FIELDS_ACCESS[i.get_name()][0], idx, self.method)

            # invoke
            elif (op_value >= 0x6e and op_value <= 0x72) or (op_value >= 0x74 and op_value <= 0x78):
                idx_meth = i.get_ref_kind()
                method_info = self.__vm.get_cm_method(idx_meth)
                if self.tainted_packages != None:
                    self.tainted_packages.push_info(method_info[0], TAINTED_PACKAGE_CALL, idx, self.method, idx_meth)

            # new_instance
            elif op_value == 0x22:
                idx_type = i.get_ref_kind()
                type_info = self.__vm.get_cm_type(idx_type)
                if self.tainted_packages != None:
                    self.tainted_packages.push_info(type_info, TAINTED_PACKAGE_CREATE, idx, self.method, None)

            # const-string
            elif (op_value >= 0x1a and op_value <= 0x1b):
                string_name = self.__vm.get_cm_string(i.get_ref_kind())
                if self.tainted_variables != None:
                    self.tainted_variables.push_info(TAINTED_STRING, string_name, "R", idx, self.method)

            elif op_value == 0x26 or (op_value >= 0x2b and op_value <= 0x2c):
                code = self.method.get_code().get_bc()
                self.special_ins[idx] = code.get_ins_off(idx + i.get_ref_off() * 2)
      except:
        pass

    def get_special_ins(self, idx):
        """
            Return the associated instruction to a specific instruction (for example a packed/sparse switch)

            :param idx: the index of the instruction

            :rtype: None or an Instruction
        """
        try:
            return self.special_ins[idx]
        except:
            return None

    def get_exception_analysis(self):
        return self.exception_analysis

    def set_exception_analysis(self, exception_analysis):
        self.exception_analysis = exception_analysis

TAINTED_LOCAL_VARIABLE = 0
TAINTED_FIELD = 1
TAINTED_STRING = 2

class PathVar(object):
  def __init__(self, access, idx, dst_idx, info_obj):
    self.access_flag = access
    self.idx = idx
    self.dst_idx = dst_idx
    self.info_obj = info_obj

  def get_var_info(self):
    return self.info_obj.get_info()

  def get_access_flag(self):
    return self.access_flag

  def get_src(self, cm):
    method = cm.get_method_ref( self.idx )
    return method.get_class_name(), method.get_name(), method.get_descriptor()

  def get_dst(self, cm):
    method = cm.get_method_ref( self.dst_idx )
    return method.get_class_name(), method.get_name(), method.get_descriptor()

  def get_idx(self):
    return self.idx

class TaintedVariable(object):
    def __init__(self, var, _type):
        self.var = var
        self.type = _type

        self.paths = {}
        self.__cache = []

    def get_type(self):
        return self.type

    def get_info(self):
        if self.type == TAINTED_FIELD:
            return [ self.var[0], self.var[2], self.var[1] ]
        return self.var

    def push(self, access, idx, ref):
        m_idx = ref.get_method_idx()

        if m_idx not in self.paths:
          self.paths[ m_idx ] = []

        self.paths[ m_idx ].append( (access, idx) )

    def get_paths_access(self, mode):
        for i in self.paths:
          for j in self.paths[ i ]:
            for k, v in self.paths[ i ][ j ]:
              if k in mode:
                yield i, j, k, v

    def get_paths(self):
        if self.__cache != []:
            return self.__cache

        for i in self.paths:
          for j in self.paths[ i ]:
              self.__cache.append( [j, i] )
              #yield j, i
        return self.__cache

    def get_paths_length(self):
        return len(self.paths)

    def show_paths(self, vm):
        show_PathVariable( vm, self.get_paths() )

class TaintedVariables(object):
    def __init__(self, _vm):
        self.__vm = _vm
        self.__vars = {
           TAINTED_LOCAL_VARIABLE : {},
           TAINTED_FIELD : {},
           TAINTED_STRING : {},
        }

        self.__cache_field_by_method = {}
        self.__cache_string_by_method = {}

        self.AOSP_PERMISSIONS_MODULE = load_api_specific_resource_module("aosp_permissions", self.__vm.get_api_version())
        self.API_PERMISSION_MAPPINGS_MODULE = load_api_specific_resource_module("api_permission_mappings", self.__vm.get_api_version())

    # functions to get particulars elements
    def get_string(self, s):
        try:
            return self.__vars[ TAINTED_STRING ][ s ]
        except KeyError:
            return None

    def get_field(self, class_name, name, descriptor):
        key = class_name + descriptor + name

        try:
            return self.__vars[ TAINTED_FIELD ] [ key ]
        except KeyError:
            return None

    def toPathVariable(self, obj):
      z = []
      for i in obj.get_paths():
        access, idx = i[0]
        m_idx = i[1]

        z.append( PathVar(access, idx, m_idx, obj ) )
      return z

    # permission functions
    def get_permissions_method(self, method):
        permissions = set()

        for f, f1 in self.get_fields():
            data = "%s-%s-%s" % (f.var[0], f.var[2], f.var[1])
            if data in self.API_PERMISSION_MAPPINGS_MODULE["AOSP_PERMISSIONS_BY_FIELDS"].keys():
                for path in f.get_paths():
                    #access, idx = path[0]
                    m_idx = path[1]
                    if m_idx == method.get_idx():
                        permissions.update(self.API_PERMISSION_MAPPINGS_MODULE["AOSP_PERMISSIONS_BY_FIELDS"][data])

        return permissions

    def get_permissions(self, permissions_needed):
        """
            @param permissions_needed : a list of restricted permissions to get ([] returns all permissions)

            @rtype : a dictionnary of permissions' paths
        """
        permissions = {}

        pn = set(permissions_needed)
        if permissions_needed == []:
            pn = set(self.AOSP_PERMISSIONS_MODULE["AOSP_PERMISSIONS"].keys())

        for f, _ in self.get_fields():
            data = "%s-%s-%s" % (f.var[0], f.var[2], f.var[1])
            if data in self.API_PERMISSION_MAPPINGS_MODULE["AOSP_PERMISSIONS_BY_FIELDS"].keys():
                perm_intersection = pn.intersection(self.API_PERMISSION_MAPPINGS_MODULE["AOSP_PERMISSIONS_BY_FIELDS"][data])
                for p in perm_intersection:
                    try:
                        permissions[p].extend(self.toPathVariable(f))
                    except KeyError:
                        permissions[p] = []
                        permissions[p].extend(self.toPathVariable(f))

        return permissions

    # global functions
    def get_strings(self):
        for i in self.__vars[ TAINTED_STRING ]:
            yield self.__vars[ TAINTED_STRING ][ i ], i

    def get_fields(self):
        for i in self.__vars[ TAINTED_FIELD ]:
            yield self.__vars[ TAINTED_FIELD ][ i ], i

    # specifics functions
    def get_strings_by_method(self, method):
        z = {}

        try:
            for i in self.__cache_string_by_method[ method.get_method_idx() ]:
                z[ i ] = []
                for j in i.get_paths():
                    if method.get_method_idx() == j[1]:
                        z[i].append( j[0] )

            return z
        except:
            return z


    def get_fields_by_method(self, method):
        z = {}

        try:
            for i in self.__cache_field_by_method[ method.get_method_idx() ]:
                z[ i ] = []
                for j in i.get_paths():
                    if method.get_method_idx() == j[1]:
                        z[i].append( j[0] )
            return z
        except:
            return z

    def add(self, var, _type, _method=None):
        if _type == TAINTED_FIELD:
            key = var[0] + var[1] + var[2]
            if key not in self.__vars[ TAINTED_FIELD ]:
                self.__vars[ TAINTED_FIELD ][ key ] = TaintedVariable( var, _type )
        elif _type == TAINTED_STRING:
            if var not in self.__vars[ TAINTED_STRING ]:
                self.__vars[ TAINTED_STRING ][ var ] = TaintedVariable( var, _type )
        elif _type == TAINTED_LOCAL_VARIABLE:
            if _method not in self.__vars[ TAINTED_LOCAL_VARIABLE ]:
                self.__vars[ TAINTED_LOCAL_VARIABLE ][ _method ] = {}

            if var not in self.__vars[ TAINTED_LOCAL_VARIABLE ][ _method ]:
                self.__vars[ TAINTED_LOCAL_VARIABLE ][ _method ][ var ] = TaintedVariable( var, _type )

    def push_info(self, _type, var, access, idx, ref):
        if _type == TAINTED_FIELD:
            self.add( var, _type )
            key = var[0] + var[1] + var[2]
            self.__vars[ _type ][ key ].push( access, idx, ref )

            method_idx = ref.get_method_idx()
            if method_idx not in self.__cache_field_by_method:
                self.__cache_field_by_method[ method_idx ] = set()

            self.__cache_field_by_method[ method_idx ].add( self.__vars[ TAINTED_FIELD ][ key ] )


        elif _type == TAINTED_STRING:
            self.add( var, _type )
            self.__vars[ _type ][ var ].push( access, idx, ref )

            method_idx = ref.get_method_idx()

            if method_idx not in self.__cache_string_by_method:
                self.__cache_string_by_method[ method_idx ] = set()

            self.__cache_string_by_method[ method_idx ].add( self.__vars[ TAINTED_STRING ][ var ] )

TAINTED_PACKAGE_CREATE = 0
TAINTED_PACKAGE_CALL = 1

TAINTED_PACKAGE = {
   TAINTED_PACKAGE_CREATE : "C",
   TAINTED_PACKAGE_CALL : "M"
}
def show_Path(vm, path):
  cm = vm.get_class_manager()

  if isinstance(path, PathVar):
    dst_class_name, dst_method_name, dst_descriptor =  path.get_dst( cm )
    info_var = path.get_var_info()
    print "%s %s (0x%x) ---> %s->%s%s" % (path.get_access_flag(),
                                          info_var,
                                          path.get_idx(),
                                          dst_class_name,
                                          dst_method_name,
                                          dst_descriptor)
  else:
    if path.get_access_flag() == TAINTED_PACKAGE_CALL:
      src_class_name, src_method_name, src_descriptor =  path.get_src( cm )
      dst_class_name, dst_method_name, dst_descriptor =  path.get_dst( cm )

      print "%d %s->%s%s (0x%x) ---> %s->%s%s" % (path.get_access_flag(),
                                                  src_class_name,
                                                  src_method_name,
                                                  src_descriptor,
                                                  path.get_idx(),
                                                  dst_class_name,
                                                  dst_method_name,
                                                  dst_descriptor)
    else:
      src_class_name, src_method_name, src_descriptor =  path.get_src( cm )
      print "%d %s->%s%s (0x%x)" % (path.get_access_flag(),
                                    src_class_name,
                                    src_method_name,
                                    src_descriptor,
                                    path.get_idx())

def get_Path(vm, path):
  x = {}
  cm = vm.get_class_manager()

  if isinstance(path, PathVar):
    dst_class_name, dst_method_name, dst_descriptor =  path.get_dst( cm )
    info_var = path.get_var_info()
    x["src"] = "%s" % info_var
    x["dst"] = "%s %s %s" % (dst_class_name, dst_method_name, dst_descriptor)
    x["idx"] = path.get_idx()

  else:
    if path.get_access_flag() == TAINTED_PACKAGE_CALL:
      src_class_name, src_method_name, src_descriptor =  path.get_src( cm )
      dst_class_name, dst_method_name, dst_descriptor =  path.get_dst( cm )

      x["src"] = "%s %s %s" % (src_class_name, src_method_name, src_descriptor)
      x["dst"] = "%s %s %s" % (dst_class_name, dst_method_name, dst_descriptor)
    else:
      src_class_name, src_method_name, src_descriptor =  path.get_src( cm )
      x["src"] = "%s %s %s" % (src_class_name, src_method_name, src_descriptor)

    x["idx"] = path.get_idx()

  return x


def show_Paths(vm, paths):
    """
        Show paths of packages
        :param vm: the object which represents the dex file
        :param paths: a list of :class:`PathP` objects
    """
    for path in paths:
        show_Path( vm, path )


def get_Paths(vm, paths):
    """
        Return paths of packages
        :param vm: the object which represents the dex file
        :param paths: a list of :class:`PathP` objects
    """
    full_paths = []
    for path in paths:
        full_paths.append(get_Path( vm, path ))
    return full_paths


def show_PathVariable(vm, paths):
    for path in paths:
      access, idx = path[0]
      m_idx = path[1]
      method = vm.get_cm_method(m_idx)
      print "%s %x %s->%s %s" % (access, idx, method[0], method[1], method[2][0] + method[2][1])


class PathP(object):
  def __init__(self, access, idx, src_idx, dst_idx):
    self.access_flag = access
    self.idx = idx
    self.src_idx = src_idx
    self.dst_idx = dst_idx

  def get_access_flag(self):
    return self.access_flag

  def get_dst(self, cm):
    method = cm.get_method_ref(self.dst_idx)
    return method.get_class_name(), method.get_name(), method.get_descriptor()

  def get_src(self, cm):
    method = cm.get_method_ref(self.src_idx)
    return method.get_class_name(), method.get_name(), method.get_descriptor()

  def get_idx(self):
    return self.idx

  def get_src_idx(self):
    return self.src_idx

  def get_dst_idx(self):
    return self.dst_idx


class TaintedPackage(object):
    def __init__(self, vm, name):
        self.vm = vm
        self.name = name
        self.paths = {TAINTED_PACKAGE_CREATE : [], TAINTED_PACKAGE_CALL : []}

    def get_name(self):
        return self.name

    def gets(self):
        return self.paths

    def push(self, access, idx, src_idx, dst_idx):
        p = PathP( access, idx, src_idx, dst_idx )
        self.paths[ access ].append( p )
        return p

    def get_objects_paths(self):
        return self.paths[ TAINTED_PACKAGE_CREATE ]

    def search_method(self, name, descriptor):
        """
            @param name : a regexp for the name of the method
            @param descriptor : a regexp for the descriptor of the method

            @rtype : a list of called paths
        """
        l = []
        m_name = re.compile(name)
        m_descriptor = re.compile(descriptor)

        for path in self.paths[ TAINTED_PACKAGE_CALL ]:
            _, dst_name, dst_descriptor = path.get_dst(self.vm.get_class_manager())

            if m_name.match( dst_name ) != None and m_descriptor.match( dst_descriptor ) != None:
                l.append( path )
        return l

    def get_method(self, name, descriptor):
        l = []
        for path in self.paths[ TAINTED_PACKAGE_CALL ]:
            if path.get_name() == name and path.get_descriptor() == descriptor:
                l.append( path )
        return l

    def get_paths(self):
        for i in self.paths:
            for j in self.paths[ i ]:
                yield j

    def get_paths_length(self):
        x = 0
        for i in self.paths:
            x += len(self.paths[ i ])
        return x

    def get_methods(self):
        return [path for path in self.paths[TAINTED_PACKAGE_CALL]]

    def get_new(self):
        return [path for path in self.paths[TAINTED_PACKAGE_CREATE]]

    def show(self):
        cm = self.vm.get_class_manager()
        print self.get_name()
        for _type in self.paths:
            print "\t -->", _type
            if _type == TAINTED_PACKAGE_CALL:
                for path in self.paths[_type]:
                    print "\t\t => %s <-- %x in %s" % (path.get_dst(cm), path.get_idx(), path.get_src(cm))
            else:
                for path in self.paths[_type]:
                    print "\t\t => %x in %s" % (path.get_idx(), path.get_src(cm))

def show_Permissions(dx):
    """
        Show where permissions are used in a specific application
        :param dx : the analysis virtual machine
        :type dx: a :class:`VMAnalysis` object
    """
    p = dx.get_permissions( [] )

    for i in p:
        print i, ":"
        for j in p[i]:
            show_Path( dx.get_vm(), j )

def show_DynCode(dx):
    """
        Show where dynamic code is used
        :param dx : the analysis virtual machine
        :type dx: a :class:`VMAnalysis` object
    """
    paths = []
    paths.extend(dx.get_tainted_packages().search_methods("Ldalvik/system/BaseDexClassLoader;",
                                                "<init>",
                                                "."))

    paths.extend(dx.get_tainted_packages().search_methods("Ldalvik/system/PathClassLoader;",
                                                "<init>",
                                                "."))

    paths.extend(dx.get_tainted_packages().search_methods("Ldalvik/system/DexClassLoader;",
                                                "<init>",
                                                "."))

    paths.extend(dx.get_tainted_packages().search_methods("Ldalvik/system/DexFile;",
                                                "<init>",
                                                "."))

    paths.extend(dx.get_tainted_packages().search_methods("Ldalvik/system/DexFile;",
                                                "loadDex",
                                                "."))
    show_Paths( dx.get_vm(), paths )


def show_NativeMethods(dx):
    """
        Show the native methods
        :param dx : the analysis virtual machine
        :type dx: a :class:`VMAnalysis` object
    """
    print get_NativeMethods(dx)


def show_ReflectionCode(dx):
    """
        Show the reflection code
        :param dx : the analysis virtual machine
        :type dx: a :class:`VMAnalysis` object
    """
    paths = dx.get_tainted_packages().search_methods("Ljava/lang/reflect/Method;", ".", ".")
    show_Paths(dx.get_vm(), paths)


def get_NativeMethods(dx):
    """
        Return the native methods
        :param dx : the analysis virtual machine
        :type dx: a :class:`VMAnalysis` object
        :rtype: [tuple]
    """
    d = dx.get_vm()
    native_methods = []
    for i in d.get_methods():
        if i.get_access_flags() & 0x100:
            native_methods.append(
                (i.get_class_name(), i.get_name(), i.get_descriptor()))
    return native_methods


def get_ReflectionCode(dx):
    """
        Return the reflection code
        :param dx : the analysis virtual machine
        :type dx: a :class:`VMAnalysis` object
        :rtype: [dict]
    """
    paths = dx.get_tainted_packages().search_methods(
        "Ljava/lang/reflect/Method;", ".", ".")
    return get_Paths(dx.get_vm(), paths)


def is_crypto_code(dx):
    """
        Crypto code is present ?
        :param dx : the analysis virtual machine
        :type dx: a :class:`VMAnalysis` object
        :rtype: boolean
    """
    if dx.get_tainted_packages().search_methods("Ljavax/crypto/.",
                                                ".",
                                                "."):
        return True

    if dx.get_tainted_packages().search_methods("Ljava/security/spec/.",
                                                ".",
                                                "."):
        return True

    return False


def is_dyn_code(dx):
    """
        Dalvik Dynamic code loading is present ?
        :param dx : the analysis virtual machine
        :type dx: a :class:`VMAnalysis` object
        :rtype: boolean
    """
    if dx.get_tainted_packages().search_methods("Ldalvik/system/BaseDexClassLoader;",
                                                "<init>",
                                                "."):
        return True

    if dx.get_tainted_packages().search_methods("Ldalvik/system/PathClassLoader;",
                                                "<init>",
                                                "."):
        return True

    if dx.get_tainted_packages().search_methods("Ldalvik/system/DexClassLoader;",
                                                "<init>",
                                                "."):
        return True

    if dx.get_tainted_packages().search_methods("Ldalvik/system/DexFile;",
                                                "<init>",
                                                "."):
        return True

    if dx.get_tainted_packages().search_methods("Ldalvik/system/DexFile;",
                                                "loadDex",
                                                "."):
        return True

    return False


def is_reflection_code(dx):
    """
        Reflection is present ?
        :param dx : the analysis virtual machine
        :type dx: a :class:`VMAnalysis` object
        :rtype: boolean
    """
    if dx.get_tainted_packages().search_methods("Ljava/lang/reflect/Method;",
                                                ".",
                                                "."):
        return True

    if dx.get_tainted_packages().search_methods("Ljava/lang/reflect/Field;",
                                                ".",
                                                "."):
        return True

    if dx.get_tainted_packages().search_methods("Ljava/lang/Class;",
                                                "forName",
                                                "."):
        return True

    return False


def is_native_code(dx):
    """
        Native code is present ?
        :param dx : the analysis virtual machine
        :type dx: a :class:`VMAnalysis` object
        :rtype: boolean
    """
    if dx.get_tainted_packages().search_methods("Ljava/lang/System;",
                                                "load.",
                                                "."):
        return True

    if dx.get_tainted_packages().search_methods("Ljava/lang/Runtime;",
                                                "load.",
                                                "."):
        return True

    return False


class TaintedPackages(object):
    def __init__(self, _vm):
        self.__vm = _vm
        self.__packages = {}
        self.__methods = {}

        self.AOSP_PERMISSIONS_MODULE = load_api_specific_resource_module("aosp_permissions", self.__vm.get_api_version())
        self.API_PERMISSION_MAPPINGS_MODULE = load_api_specific_resource_module("api_permission_mappings", self.__vm.get_api_version())

    def _add_pkg(self, name):
        if name not in self.__packages:
            self.__packages[ name ] = TaintedPackage( self.__vm, name )

    #self.context.get_tainted_packages().push_info( method_info[0], TAINTED_PACKAGE_CALL, idx, self, self.method, method_info[1], method_info[2][0] + method_info[2][1] )
    def push_info(self, class_name, access, idx, method, idx_method):
        self._add_pkg( class_name )
        p = self.__packages[ class_name ].push( access, idx, method.get_method_idx(), idx_method )

        try:
            self.__methods[ method ][ class_name ].append( p )
        except:
            try:
                self.__methods[ method ][ class_name ] = []
            except:
                self.__methods[ method ] = {}
                self.__methods[ method ][ class_name ] = []

            self.__methods[ method ][ class_name ].append( p )

    def get_packages_by_method(self, method):
        try:
            return self.__methods[method]
        except KeyError:
            return {}

    def get_package(self, name):
        return self.__packages[name]

    def get_packages_by_bb(self, bb):
        """
            :rtype: return a list of packaged used in a basic block
        """
        l = []
        for i in self.__packages:
            paths = self.__packages[i].gets()
            for j in paths:
                for k in paths[j]:
                    if k.get_bb() == bb:
                        l.append( (i, k.get_access_flag(), k.get_idx(), k.get_method()) )

        return l

    def get_packages(self):
        for i in self.__packages:
            yield self.__packages[i], i

    def get_internal_packages_from_package(self, package):
        classes = self.__vm.get_classes_names()
        l = []
        for m, _ in self.get_packages():
            paths = m.get_methods()
            for j in paths:
                src_class_name, _, _ = j.get_src(self.__vm.get_class_manager())
                dst_class_name, _, _ = j.get_dst(self.__vm.get_class_manager())

                if src_class_name == package and dst_class_name in classes:
                    l.append(j)
        return l

    def get_internal_packages(self):
        """
            :rtype: return a list of the internal packages called in the application
        """
        classes = self.__vm.get_classes_names()
        l = []
        for m, _ in self.get_packages():
            paths = m.get_methods()
            for j in paths:
                if j.get_access_flag() == TAINTED_PACKAGE_CALL:
                  dst_class_name, _, _ = j.get_dst(self.__vm.get_class_manager())
                  if dst_class_name in classes and m.get_name() in classes:
                    l.append(j)
        return l

    def get_internal_new_packages(self):
        """
            :rtype: return a list of the internal packages created in the application
        """
        classes = self.__vm.get_classes_names()
        l = {}
        for m, _ in self.get_packages():
            paths = m.get_new()
            for j in paths:
                src_class_name, _, _ = j.get_src(self.__vm.get_class_manager())
                if src_class_name in classes and m.get_name() in classes:
                    if j.get_access_flag() == TAINTED_PACKAGE_CREATE:
                        try:
                            l[m.get_name()].append(j)
                        except:
                            l[m.get_name()] = []
                            l[m.get_name()].append(j)
        return l

    def get_external_packages(self):
        """
            :rtype: return a list of the external packages called in the application
        """
        classes = self.__vm.get_classes_names()
        l = []
        for m, _ in self.get_packages():
            paths = m.get_methods()
            for j in paths:
                src_class_name, _, _ = j.get_src(self.__vm.get_class_manager())
                dst_class_name, _, _ = j.get_dst(self.__vm.get_class_manager())
                if src_class_name in classes and dst_class_name not in classes:
                    if j.get_access_flag() == TAINTED_PACKAGE_CALL:
                        l.append(j)
        return l

    def search_packages(self, package_name):
        """
            :param package_name: a regexp for the name of the package

            :rtype: a list of called packages' paths
        """
        ex = re.compile(package_name)

        l = []
        for m, _ in self.get_packages():
            if ex.search(m.get_name()) != None:
                l.extend(m.get_methods())
        return l

    def search_unique_packages(self, package_name):
        """
            :param package_name: a regexp for the name of the package
        """
        ex = re.compile( package_name )

        l = []
        d = {}
        for m, _ in self.get_packages():
            if ex.match( m.get_info() ) != None:
                for path in m.get_methods():
                    try:
                        d[ path.get_class_name() + path.get_name() + path.get_descriptor() ] += 1
                    except KeyError:
                        d[ path.get_class_name() + path.get_name() + path.get_descriptor() ] = 0
                        l.append( [ path.get_class_name(), path.get_name(), path.get_descriptor() ] )
        return l, d

    def search_methods(self, class_name, name, descriptor, re_expr=True):
        """
            @param class_name : a regexp for the class name of the method (the package)
            @param name : a regexp for the name of the method
            @param descriptor : a regexp for the descriptor of the method

            @rtype : a list of called methods' paths
        """
        l = []
        if re_expr == True:
            ex = re.compile( class_name )

            for m, _ in self.get_packages():
                if ex.search( m.get_name() ) != None:
                    l.extend( m.search_method( name, descriptor ) )

        return l

    def search_objects(self, class_name):
        """
            @param class_name : a regexp for the class name

            @rtype : a list of created objects' paths
        """
        ex = re.compile( class_name )
        l = []

        for m, _ in self.get_packages():
            if ex.search( m.get_name() ) != None:
                l.extend( m.get_objects_paths() )

        return l

    def search_crypto_packages(self):
        """
            @rtype : a list of called crypto packages
        """
        return self.search_packages( "Ljavax/crypto/" )

    def search_telephony_packages(self):
        """
            @rtype : a list of called telephony packages
        """
        return self.search_packages( "Landroid/telephony/" )

    def search_net_packages(self):
        """
            @rtype : a list of called net packages
        """
        return self.search_packages( "Landroid/net/" )

    def get_method(self, class_name, name, descriptor):
        try:
            return self.__packages[ class_name ].get_method( name, descriptor )
        except KeyError:
            return []

    def get_permissions_method(self, method):
        permissions = set()
        for m, _ in self.get_packages():
            paths = m.get_methods()
            for j in paths:
                if j.get_method() == method:
                    if j.get_access_flag() == TAINTED_PACKAGE_CALL:
                        dst_class_name, dst_method_name, dst_descriptor = j.get_dst( self.__vm.get_class_manager() )
                        data = "%s-%s-%s" % (dst_class_name, dst_method_name, dst_descriptor)
                        if data in self.API_PERMISSION_MAPPINGS_MODULE["AOSP_PERMISSIONS_BY_METHODS"].keys():
                            permissions.update(self.API_PERMISSION_MAPPINGS_MODULE["AOSP_PERMISSIONS_BY_METHODS"][data])

        return permissions

    def get_permissions(self, permissions_needed):
        """
            @param permissions_needed : a list of restricted permissions to get ([] returns all permissions)
            @rtype : a dictionnary of permissions' paths
        """
        permissions = {}

        pn = set(permissions_needed)
        if permissions_needed == []:
            pn = set(self.AOSP_PERMISSIONS_MODULE["AOSP_PERMISSIONS"].keys())

        classes = self.__vm.get_classes_names()

        for m, _ in self.get_packages():
            paths = m.get_methods()
            for j in paths:
                src_class_name, src_method_name, src_descriptor = j.get_src( self.__vm.get_class_manager() )
                dst_class_name, dst_method_name, dst_descriptor = j.get_dst( self.__vm.get_class_manager() )
                if (src_class_name in classes) and (dst_class_name not in classes):
                    if j.get_access_flag() == TAINTED_PACKAGE_CALL:
                        data = "%s-%s-%s" % (dst_class_name, dst_method_name, dst_descriptor)
                        if data in self.API_PERMISSION_MAPPINGS_MODULE["AOSP_PERMISSIONS_BY_METHODS"].keys():
                            perm_intersection = pn.intersection(self.API_PERMISSION_MAPPINGS_MODULE["AOSP_PERMISSIONS_BY_METHODS"][data])
                            for p in perm_intersection:
                                try:
                                    permissions[p].append(j)
                                except KeyError:
                                    permissions[p] = []
                                    permissions[p].append(j)

        return permissions

class Enum(object):
  def __init__(self, names):
    self.names = names
    for value, name in enumerate(self.names):
      setattr(self, name.upper(), value)

  def tuples(self):
    return tuple(enumerate(self.names))

TAG_ANDROID = Enum([ 'ANDROID', 'TELEPHONY', 'SMS', 'SMSMESSAGE', 'ACCESSIBILITYSERVICE', 'ACCOUNTS',
    'ANIMATION', 'APP', 'BLUETOOTH', 'CONTENT', 'DATABASE', 'DEBUG', 'DRM', 'GESTURE',
    'GRAPHICS', 'HARDWARE', 'INPUTMETHODSERVICE', 'LOCATION', 'MEDIA', 'MTP',
    'NET', 'NFC', 'OPENGL', 'OS', 'PREFERENCE', 'PROVIDER', 'RENDERSCRIPT',
    'SAX', 'SECURITY', 'SERVICE', 'SPEECH', 'SUPPORT', 'TEST', 'TEXT', 'UTIL',
    'VIEW', 'WEBKIT', 'WIDGET', 'DALVIK_BYTECODE', 'DALVIK_SYSTEM', 'JAVA_REFLECTION'])

TAG_REVERSE_ANDROID = dict((i[0], i[1]) for i in TAG_ANDROID.tuples())

TAGS_ANDROID = { TAG_ANDROID.ANDROID :                  [ 0, "Landroid" ],
                 TAG_ANDROID.TELEPHONY :                [ 0, "Landroid/telephony"],
                 TAG_ANDROID.SMS :                      [ 0, "Landroid/telephony/SmsManager"],
                 TAG_ANDROID.SMSMESSAGE :               [ 0, "Landroid/telephony/SmsMessage"],
                 TAG_ANDROID.DEBUG :                    [ 0, "Landroid/os/Debug"],
                 TAG_ANDROID.ACCESSIBILITYSERVICE :     [ 0, "Landroid/accessibilityservice" ],
                 TAG_ANDROID.ACCOUNTS :                 [ 0, "Landroid/accounts" ],
                 TAG_ANDROID.ANIMATION :                [ 0, "Landroid/animation" ],
                 TAG_ANDROID.APP :                      [ 0, "Landroid/app" ],
                 TAG_ANDROID.BLUETOOTH :                [ 0, "Landroid/bluetooth" ],
                 TAG_ANDROID.CONTENT :                  [ 0, "Landroid/content" ],
                 TAG_ANDROID.DATABASE :                 [ 0, "Landroid/database" ],
                 TAG_ANDROID.DRM :                      [ 0, "Landroid/drm" ],
                 TAG_ANDROID.GESTURE :                  [ 0, "Landroid/gesture" ],
                 TAG_ANDROID.GRAPHICS :                 [ 0, "Landroid/graphics" ],
                 TAG_ANDROID.HARDWARE :                 [ 0, "Landroid/hardware" ],
                 TAG_ANDROID.INPUTMETHODSERVICE :       [ 0, "Landroid/inputmethodservice" ],
                 TAG_ANDROID.LOCATION :                 [ 0, "Landroid/location" ],
                 TAG_ANDROID.MEDIA :                    [ 0, "Landroid/media" ],
                 TAG_ANDROID.MTP :                      [ 0, "Landroid/mtp" ],
                 TAG_ANDROID.NET :                      [ 0, "Landroid/net" ],
                 TAG_ANDROID.NFC :                      [ 0, "Landroid/nfc" ],
                 TAG_ANDROID.OPENGL :                   [ 0, "Landroid/opengl" ],
                 TAG_ANDROID.OS :                       [ 0, "Landroid/os" ],
                 TAG_ANDROID.PREFERENCE :               [ 0, "Landroid/preference" ],
                 TAG_ANDROID.PROVIDER :                 [ 0, "Landroid/provider" ],
                 TAG_ANDROID.RENDERSCRIPT :             [ 0, "Landroid/renderscript" ],
                 TAG_ANDROID.SAX :                      [ 0, "Landroid/sax" ],
                 TAG_ANDROID.SECURITY :                 [ 0, "Landroid/security" ],
                 TAG_ANDROID.SERVICE :                  [ 0, "Landroid/service" ],
                 TAG_ANDROID.SPEECH :                   [ 0, "Landroid/speech" ],
                 TAG_ANDROID.SUPPORT :                  [ 0, "Landroid/support" ],
                 TAG_ANDROID.TEST :                     [ 0, "Landroid/test" ],
                 TAG_ANDROID.TEXT :                     [ 0, "Landroid/text" ],
                 TAG_ANDROID.UTIL :                     [ 0, "Landroid/util" ],
                 TAG_ANDROID.VIEW :                     [ 0, "Landroid/view" ],
                 TAG_ANDROID.WEBKIT :                   [ 0, "Landroid/webkit" ],
                 TAG_ANDROID.WIDGET :                   [ 0, "Landroid/widget" ],
                 TAG_ANDROID.DALVIK_BYTECODE :          [ 0, "Ldalvik/bytecode" ],
                 TAG_ANDROID.DALVIK_SYSTEM :            [ 0, "Ldalvik/system" ],

                 TAG_ANDROID.JAVA_REFLECTION :          [ 0, "Ljava/lang/reflect"],
}

class Tags(object):
  """
      Handle specific tags

      :param patterns:
      :params reverse:
  """
  def __init__(self, patterns=TAGS_ANDROID, reverse=TAG_REVERSE_ANDROID):
    self.tags = set()

    self.patterns = patterns
    self.reverse = TAG_REVERSE_ANDROID

    for i in self.patterns:
      self.patterns[i][1] = re.compile(self.patterns[i][1])

  def emit(self, method):
    for i in self.patterns:
      if self.patterns[i][0] == 0:
        if self.patterns[i][1].search( method.get_class() ) != None:
          self.tags.add( i )

  def emit_by_classname(self, classname):
    for i in self.patterns:
      if self.patterns[i][0] == 0:
        if self.patterns[i][1].search( classname ) != None:
          self.tags.add( i )

  def get_list(self):
    return [ self.reverse[ i ] for i in self.tags ]

  def __contains__(self, key):
    return key in self.tags

  def __str__(self):
    return str([ self.reverse[ i ] for i in self.tags ])


  def empty(self):
    return self.tags == set()


class BasicBlocks(object):
    """
        This class represents all basic blocks of a method
    """
    def __init__(self, _vm, tv):
        self.__vm = _vm
        self.tainted = tv

        self.bb = []

    def push(self, bb):
        self.bb.append(bb)

    def pop(self, idx):
        return self.bb.pop(idx)

    def get_basic_block(self, idx):
        for i in self.bb:
            if idx >= i.get_start() and idx < i.get_end():
                return i
        return None

    def get_tainted_integers(self):
        try:
          return self.tainted.get_tainted_integers()
        except:
          return None

    def get_tainted_packages(self):
        try:
          return self.tainted.get_tainted_packages()
        except:
          return None

    def get_tainted_variables(self):
        try:
          return self.tainted.get_tainted_variables()
        except:
          return None

    def get(self):
        """
            :rtype: return each basic block (:class:`DVMBasicBlock` object)
        """
        for i in self.bb:
            yield i

    def gets(self):
        """
            :rtype: a list of basic blocks (:class:`DVMBasicBlock` objects)
        """
        return self.bb

    def get_basic_block_pos(self, idx):
        return self.bb[idx]


class ExceptionAnalysis(object):
    def __init__(self, exception, bb):
        self.start = exception[0]
        self.end = exception[1]

        self.exceptions = exception[2:]

        for i in self.exceptions:
            i.append(bb.get_basic_block(i[1]))

    def show_buff(self):
        buff = "%x:%x\n" % (self.start, self.end)

        for i in self.exceptions:
            if i[2] == None:
                buff += "\t(%s -> %x %s)\n" % (i[0], i[1], i[2])
            else:
                buff += "\t(%s -> %x %s)\n" % (i[0], i[1], i[2].get_name())

        return buff[:-1]

    def get(self):
        d = {"start": self.start, "end": self.end, "list": []}

        for i in self.exceptions:
            d["list"].append({"name": i[0], "idx": i[1], "bb": i[2].get_name()})

        return d


class Exceptions(object):
    def __init__(self, _vm):
        self.__vm = _vm
        self.exceptions = []

    def add(self, exceptions, basic_blocks):
        for i in exceptions:
            self.exceptions.append( ExceptionAnalysis( i, basic_blocks ) )

    def get_exception(self, addr_start, addr_end):
        for i in self.exceptions:
#            print hex(i.start), hex(i.end), hex(addr_start), hex(addr_end), i.start >= addr_start and i.end <= addr_end, addr_end <= i.end and addr_start >= i.start
            if i.start >= addr_start and i.end <= addr_end:
                return i

            elif addr_end <= i.end and addr_start >= i.start:
                return i

        return None

    def gets(self):
        return self.exceptions

    def get(self):
        for i in self.exceptions:
            yield i

BO = { "BasicOPCODES" : dvm.BRANCH_DVM_OPCODES, "BasicClass" : DVMBasicBlock, "Dnext" : dvm.determineNext, "Dexception" : dvm.determineException }

BO["BasicOPCODES_H"] = []
for i in BO["BasicOPCODES"]:
  BO["BasicOPCODES_H"].append( re.compile( i ) )


class MethodAnalysis(object):
    """
        This class analyses in details a method of a class/dex file

        :param vm: the object which represent the dex file
        :param method: the original method
        :param tv: a virtual object to get access to tainted information
        :type vm: a :class:`DalvikVMFormat` object
        :type method: a :class:`EncodedMethod` object
    """
    def __init__(self, vm, method, tv):
        self.__vm = vm
        self.method = method

        self.tainted = tv

        self.basic_blocks = BasicBlocks(self.__vm, self.tainted)
        self.exceptions = Exceptions(self.__vm)

        code = self.method.get_code()
        if code == None:
            return

        current_basic = BO["BasicClass"](0, self.__vm, self.method, self.basic_blocks)
        self.basic_blocks.push(current_basic)

        ##########################################################

        bc = code.get_bc()
        l = []
        h = {}
        idx = 0

        debug("Parsing instructions")
        instructions = [i for i in bc.get_instructions()]
        for i in instructions:
            for j in BO["BasicOPCODES_H"]:
                if j.match(i.get_name()) != None:
                    v = BO["Dnext"](i, idx, self.method)
                    h[ idx ] = v
                    l.extend(v)
                    break

            idx += i.get_length()

        debug("Parsing exceptions")
        excepts = BO["Dexception"]( self.__vm, self.method )
        for i in excepts:
            l.extend( [i[0]] )
            for handler in i[2:]:
                l.append( handler[1] )

        debug("Creating basic blocks in %s" % self.method)
        idx = 0
        for i in instructions:
            # index is a destination
            if idx in l:
                if current_basic.get_nb_instructions() != 0:
                    current_basic = BO["BasicClass"](current_basic.get_end(), self.__vm, self.method, self.basic_blocks)
                    self.basic_blocks.push(current_basic)

            current_basic.push(i)

            # index is a branch instruction
            if idx in h:
                current_basic = BO["BasicClass"]( current_basic.get_end(), self.__vm, self.method, self.basic_blocks )
                self.basic_blocks.push( current_basic )

            idx += i.get_length()

        if current_basic.get_nb_instructions() == 0:
            self.basic_blocks.pop(-1)

        debug("Settings basic blocks childs")

        for i in self.basic_blocks.get():
            try:
                i.set_childs( h[ i.end - i.get_last_length() ] )
            except KeyError:
                i.set_childs( [] )

        debug("Creating exceptions")

        # Create exceptions
        self.exceptions.add(excepts, self.basic_blocks)

        for i in self.basic_blocks.get():
            # setup exception by basic block
            i.set_exception_analysis(self.exceptions.get_exception( i.start, i.end - 1 ))

        del instructions
        del h, l

    def get_basic_blocks(self):
        """
            :rtype: a :class:`BasicBlocks` object
        """
        return self.basic_blocks

    def get_length(self):
        """
            :rtype: an integer which is the length of the code
        """
        return self.get_code().get_length()

    def get_vm(self):
        return self.__vm

    def get_method(self):
        return self.method

    def get_local_variables(self):
        return self.tainted.get_tainted_variables().get_local_variables( self.method )

    def show(self):
        print "METHOD", self.method.get_class_name(), self.method.get_name(), self.method.get_descriptor()

        for i in self.basic_blocks.get():
            print "\t", i
            i.show()
            print ""

    def show_methods(self):
        print "\t #METHODS :"
        for i in self.__bb:
            methods = i.get_methods()
            for method in methods:
                print "\t\t-->", method.get_class_name(), method.get_name(), method.get_descriptor()
                for context in methods[method]:
                    print "\t\t\t |---|", context.details

    def create_tags(self):
      """
          Create the tags for the method
      """
      self.tags = Tags()
      for i in self.tainted.get_tainted_packages().get_packages_by_method( self.method ):
        self.tags.emit_by_classname( i )

    def get_tags(self):
      """
          Return the tags of the method

          :rtype: a :class:`Tags` object
      """
      return self.tags

SIGNATURE_L0_0 = "L0_0"
SIGNATURE_L0_1 = "L0_1"
SIGNATURE_L0_2 = "L0_2"
SIGNATURE_L0_3 = "L0_3"
SIGNATURE_L0_4 = "L0_4"
SIGNATURE_L0_5 = "L0_5"
SIGNATURE_L0_6 = "L0_6"
SIGNATURE_L0_0_L1 = "L0_0:L1"
SIGNATURE_L0_1_L1 = "L0_1:L1"
SIGNATURE_L0_2_L1 = "L0_2:L1"
SIGNATURE_L0_3_L1 = "L0_3:L1"
SIGNATURE_L0_4_L1 = "L0_4:L1"
SIGNATURE_L0_5_L1 = "L0_5:L1"
SIGNATURE_L0_0_L2 = "L0_0:L2"
SIGNATURE_L0_0_L3 = "L0_0:L3"
SIGNATURE_HEX = "hex"
SIGNATURE_SEQUENCE_BB = "sequencebb"

SIGNATURES = {
                SIGNATURE_L0_0 : { "type" : 0 },
                SIGNATURE_L0_1 : { "type" : 1 },
                SIGNATURE_L0_2 : { "type" : 2, "arguments" : ["Landroid"] },
                SIGNATURE_L0_3 : { "type" : 2, "arguments" : ["Ljava"] },
                SIGNATURE_L0_4 : { "type" : 2, "arguments" : ["Landroid", "Ljava"] },
                SIGNATURE_L0_5 : { "type" : 3, "arguments" : ["Landroid"] },
                SIGNATURE_L0_6 : { "type" : 3, "arguments" : ["Ljava"] },
                SIGNATURE_SEQUENCE_BB : {},
                SIGNATURE_HEX : {},
            }

from sign import Signature

class StringAnalysis(object):
    def __init__(self, value):
        self.value = value
        self.xreffrom = set()

    def AddXrefFrom(self, classobj, methodobj):
        #debug("Added strings xreffrom for %s to %s" % (self.value, methodobj))
        self.xreffrom.add((classobj, methodobj))

    def get_xref_from(self):
        return self.xreffrom

    def __str__(self):
        data = "XREFto for string %s in\n" % repr(self.value)
        for ref_class, ref_method in self.xreffrom:
            data += "%s:%s\n" % (ref_class.get_vm_class().get_name(), ref_method)
        return data


class MethodClassAnalysis(object):
    def __init__(self, method):
        self.method = method
        self.xrefto = set()
        self.xreffrom = set()

    def AddXrefTo(self, classobj, methodobj):
        #debug("Added method xrefto for %s [%s] to %s" % (self.method, classobj, methodobj))
        self.xrefto.add((classobj, methodobj))

    def AddXrefFrom(self, classobj, methodobj):
        #debug("Added method xreffrom for %s [%s] to %s" % (self.method, classobj, methodobj))
        self.xreffrom.add((classobj, methodobj))

    def get_xref_from(self):
        return self.xreffrom

    def get_xref_to(self):
        return self.xrefto

    def __str__(self):
        data = "XREFto for %s\n" % self.method
        for ref_class, ref_method in self.xrefto:
            data += "in\n"
            data += "%s:%s\n" % (ref_class.get_vm_class().get_name(), ref_method)

        data += "XREFFrom for %s\n" % self.method
        for ref_class, ref_method in self.xreffrom:
            data += "in\n"
            data += "%s:%s\n" % (ref_class.get_vm_class().get_name(), ref_method)

        return data

class FieldClassAnalysis(object):
    def __init__(self, field):
        self.field = field
        self.xrefread = set()
        self.xrefwrite = set()

    def AddXrefRead(self, classobj, methodobj):
        #debug("Added method xrefto for %s [%s] to %s" % (self.method, classobj, methodobj))
        self.xrefread.add((classobj, methodobj))

    def AddXrefWrite(self, classobj, methodobj):
        #debug("Added method xreffrom for %s [%s] to %s" % (self.method, classobj, methodobj))
        self.xrefwrite.add((classobj, methodobj))

    def get_xref_read(self):
        return self.xrefread

    def get_xref_write(self):
        return self.xrefwrite

    def __str__(self):
        data = "XREFRead for %s\n" % self.field
        for ref_class, ref_method in self.xrefread:
            data += "in\n"
            data += "%s:%s\n" % (ref_class.get_vm_class().get_name(), ref_method)

        data += "XREFWrite for %s\n" % self.field
        for ref_class, ref_method in self.xrefwrite:
            data += "in\n"
            data += "%s:%s\n" % (ref_class.get_vm_class().get_name(), ref_method)

        return data

REF_NEW_INSTANCE = 0
REF_CLASS_USAGE = 1

class ClassAnalysis(object):
    def __init__(self, classobj):
        self._class = classobj
        self._methods = {}
        self._fields = {}

        self.xrefto = collections.defaultdict(set)
        self.xreffrom = collections.defaultdict(set)

    def get_method_analysis(self, method):
        return self._methods.get(method)

    def get_field_analysis(self, field):
        return self._fields.get(field)

    def AddFXrefRead(self, method, classobj, field):
        if field not in self._fields:
            self._fields[field] = FieldClassAnalysis(field)
        self._fields[field].AddXrefRead(classobj, method)

    def AddFXrefWrite(self, method, classobj, field):
        if field not in self._fields:
            self._fields[field] = FieldClassAnalysis(field)
        self._fields[field].AddXrefWrite(classobj, method)

    def AddMXrefTo(self, method1, classobj, method2):
        if method1 not in self._methods:
            self._methods[method1] = MethodClassAnalysis(method1)
        self._methods[method1].AddXrefTo(classobj, method2)

    def AddMXrefFrom(self, method1, classobj, method2):
        if method1 not in self._methods:
            self._methods[method1] = MethodClassAnalysis(method1)
        self._methods[method1].AddXrefFrom(classobj, method2)

    def AddXrefTo(self, ref_kind, classobj, methodobj):
        #debug("Added class xrefto for %s to %s" % (self._class.get_name(), classobj.get_vm_class().get_name()))
        self.xrefto[classobj].add((ref_kind, methodobj))

    def AddXrefFrom(self, ref_kind, classobj, methodobj):
        #debug("Added class xreffrom for %s to %s" % (self._class.get_name(), classobj.get_vm_class().get_name()))
        self.xreffrom[classobj].add((ref_kind, methodobj))

    def get_xref_from(self):
        return self.xreffrom

    def get_xref_to(self):
        return self.xrefto

    def get_vm_class(self):
        return self._class

    def __str__(self):
        data = "XREFto for %s\n" % self._class
        for ref_class in self.xrefto:
            data += str(ref_class.get_vm_class().get_name()) + " "
            data += "in\n"
            for ref_kind, ref_method in self.xrefto[ref_class]:
                data += "%d %s\n" % (ref_kind, ref_method)

            data += "\n"

        data += "XREFFrom for %s\n" % self._class
        for ref_class in self.xreffrom:
            data += str(ref_class.get_vm_class().get_name()) + " "
            data += "in\n"
            for ref_kind, ref_method in self.xreffrom[ref_class]:
                data += "%d %s\n" % (ref_kind, ref_method)

            data += "\n"

        return data

class newVMAnalysis(object):
    def __init__(self, vm):
        self.vm = vm
        self.classes = {}
        self.strings = {}

        for current_class in self.vm.get_classes():
            self.classes[current_class.get_name()] = ClassAnalysis(current_class)

    def create_xref(self):
        debug("Creating XREF/DREF")

        instances_class_name = self.classes.keys()
        external_instances = {}

        for current_class in self.vm.get_classes():
            for current_method in current_class.get_methods():
                debug("Creating XREF for %s" % current_method)

                code = current_method.get_code()
                if code == None:
                    continue

                off = 0
                bc = code.get_bc()
                for instruction in bc.get_instructions():
                    op_value = instruction.get_op_value()
                    if op_value in [0x1c, 0x22]:
                        idx_type = instruction.get_ref_kind()
                        type_info = self.vm.get_cm_type(idx_type)

                        # Internal xref related to class manipulation
                        if type_info in instances_class_name and type_info != current_class.get_name():
                            # new instance
                            if op_value == 0x22:
                                self.classes[current_class.get_name()].AddXrefTo(REF_NEW_INSTANCE, self.classes[type_info], current_method)
                                self.classes[type_info].AddXrefFrom(REF_NEW_INSTANCE, self.classes[current_class.get_name()], current_method)
                            # class reference
                            else:
                                self.classes[current_class.get_name()].AddXrefTo(REF_CLASS_USAGE, self.classes[type_info], current_method)
                                self.classes[type_info].AddXrefFrom(REF_CLASS_USAGE, self.classes[current_class.get_name()], current_method)

                    elif ((op_value >= 0x6e and op_value <= 0x72) or
                        (op_value >= 0x74 and op_value <= 0x78)):
                            idx_meth = instruction.get_ref_kind()
                            method_info = self.vm.get_cm_method(idx_meth)
                            if method_info:
                                class_info = method_info[0]

                                method_item = self.vm.get_method_descriptor(method_info[0], method_info[1], ''.join(method_info[2]))
                                if method_item:
                                    self.classes[current_class.get_name()].AddMXrefTo(current_method, self.classes[class_info], method_item)
                                    self.classes[class_info].AddMXrefFrom(method_item, self.classes[current_class.get_name()], current_method)

                                    # Internal xref related to class manipulation
                                    if class_info in instances_class_name and class_info != current_class.get_name():
                                        self.classes[current_class.get_name()].AddXrefTo(REF_CLASS_USAGE, self.classes[class_info], method_item)
                                        self.classes[class_info].AddXrefFrom(REF_CLASS_USAGE, self.classes[current_class.get_name()], current_method)

                    elif op_value >= 0x1a and op_value <= 0x1b:
                        string_value = self.vm.get_cm_string(instruction.get_ref_kind())
                        if string_value not in self.strings:
                            self.strings[string_value] = StringAnalysis(string_value)
                        self.strings[string_value].AddXrefFrom(self.classes[current_class.get_name()], current_method)

                    elif op_value >= 0x52 and op_value <= 0x6d:
                        idx_field = instruction.get_ref_kind()
                        field_info = self.vm.get_cm_field(idx_field)
                        field_item = self.vm.get_field_descriptor(field_info[0], field_info[2], field_info[1])
                        if field_item:
                            # read access to a field
                            if (op_value >= 0x52 and op_value <= 0x58) or (op_value >= 0x60 and op_value <= 0x66):
                                self.classes[current_class.get_name()].AddFXrefRead(current_method, self.classes[current_class.get_name()], field_item)
                            # write access to a field
                            else:
                                self.classes[current_class.get_name()].AddFXrefWrite(current_method, self.classes[current_class.get_name()], field_item)

                    off += instruction.get_length()

    def get_method(self, method):
        return MethodAnalysis( self.vm, method, None )

    def get_method_by_name(self, class_name, method_name, method_descriptor):
        print class_name, method_name, method_descriptor
        if class_name in self.classes:
            for method in self.classes[class_name].get_vm_class().get_methods():
                print method.get_name(), method.get_descriptor()
                if method.get_name() == method_name and method.get_descriptor() == method_descriptor:
                    return method
        return None

    def is_class_present(self, class_name):
        return class_name in self.classes

    def get_class_analysis(self, class_name):
        return self.classes.get(class_name)

    def get_strings_analysis(self):
        return self.strings

class VMAnalysis(object):
    """
       This class analyses a dex file

       :param _vm: the object which represent the dex file
       :type _vm: a :class:`DalvikVMFormat` object

       :Example:
            VMAnalysis( DalvikVMFormat( read("toto.dex", binary=False) ) )
    """
    def __init__(self, vm):
        self.vm = vm

        self.tainted_variables = TaintedVariables( self.vm )
        self.tainted_packages = TaintedPackages( self.vm )

        self.tainted = { "variables" : self.tainted_variables,
                         "packages" : self.tainted_packages,
                       }

        self.signature = None

        for i in self.vm.get_all_fields():
            self.tainted_variables.add( [ i.get_class_name(), i.get_descriptor(), i.get_name() ], TAINTED_FIELD )

        self.methods = []
        self.hmethods = {}
        self.__nmethods = {}
        for i in self.vm.get_methods():
            x = MethodAnalysis( self.vm, i, self )
            self.methods.append( x )
            self.hmethods[ i ] = x
            self.__nmethods[ i.get_name() ] = x

    def get_vm(self):
        return self.vm

    def get_method(self, method):
        """
            Return an analysis method

            :param method: a classical method object
            :type method: an :class:`EncodedMethod` object

            :rtype: a :class:`MethodAnalysis` object
        """
        return self.hmethods[ method ]

    def get_methods(self):
        """
           Return each analysis method

           :rtype: a :class:`MethodAnalysis` object
        """
        for i in self.hmethods:
            yield self.hmethods[i]

    def get_method_signature(self, method, grammar_type="", options={}, predef_sign=""):
        """
            Return a specific signature for a specific method

            :param method: a reference to method from a vm class
            :type method: a :class:`EncodedMethod` object

            :param grammar_type: the type of the signature (optional)
            :type grammar_type: string

            :param options: the options of the signature (optional)
            :param options: dict

            :param predef_sign: used a predefined signature (optional)
            :type predef_sign: string

            :rtype: a :class:`Sign` object
        """
        if self.signature == None:
          self.signature = Signature( self )

        if predef_sign != "":
            g = ""
            o = {}

            for i in predef_sign.split(":"):
                if "_" in i:
                    g += "L0:"
                    o[ "L0" ] = SIGNATURES[ i ]
                else:
                    g += i
                    g += ":"

            return self.signature.get_method( self.get_method( method ), g[:-1], o )
        else:
            return self.signature.get_method( self.get_method( method ), grammar_type, options )

    def get_permissions(self, permissions_needed):
        """
            Return the permissions used

            :param permissions_needed: a list of restricted permissions to get ([] returns all permissions)
            :type permissions_needed: list

            :rtype: a dictionnary of permissions paths
        """
        permissions = {}

        permissions.update( self.get_tainted_packages().get_permissions( permissions_needed ) )
        permissions.update( self.get_tainted_variables().get_permissions( permissions_needed ) )

        return permissions

    def get_permissions_method(self, method):
        permissions_f = self.get_tainted_packages().get_permissions_method( method )
        permissions_v = self.get_tainted_variables().get_permissions_method( method )

        all_permissions_of_method = permissions_f.union(permissions_v)

        return list(all_permissions_of_method)

    def get_tainted_variables(self):
        """
           Return the tainted variables

           :rtype: a :class:`TaintedVariables` object
        """
        return self.tainted_variables

    def get_tainted_packages(self):
        """
           Return the tainted packages

           :rtype: a :class:`TaintedPackages` object
        """
        return self.tainted_packages

    def get_tainted_fields(self):
        return self.get_tainted_variables().get_fields()

    def get_tainted_field(self, class_name, name, descriptor):
        """
           Return a specific tainted field

           :param class_name: the name of the class
           :param name: the name of the field
           :param descriptor: the descriptor of the field
           :type class_name: string
           :type name: string
           :type descriptor: string

           :rtype: a :class:`TaintedVariable` object
        """
        return self.get_tainted_variables().get_field( class_name, name, descriptor )

class uVMAnalysis(VMAnalysis):
  """
     This class analyses a dex file but on the fly (quicker !)

     :param _vm: the object which represent the dex file
     :type _vm: a :class:`DalvikVMFormat` object

     :Example:
          uVMAnalysis( DalvikVMFormat( read("toto.dex", binary=False) ) )
  """
  def __init__(self, vm):
    self.vm = vm
    self.tainted_variables = TaintedVariables( self.vm )
    self.tainted_packages = TaintedPackages( self.vm )

    self.tainted = { "variables" : self.tainted_variables,
                     "packages" : self.tainted_packages,
    }

    self.signature = None
    self.resolve = False

  def get_methods(self):
    self.resolve = True
    for i in self.vm.get_methods():
      yield MethodAnalysis(self.vm, i, self)

  def get_method(self, method):
    return MethodAnalysis( self.vm, method, None )

  def get_vm(self):
    return self.vm

  def _resolve(self):
    if self.resolve == False:
      for i in self.get_methods():
        pass

  def get_tainted_packages(self):
    self._resolve()
    return self.tainted_packages

  def get_tainted_variables(self):
        self._resolve()
        return self.tainted_variables

def is_ascii_obfuscation(vm):
    for classe in vm.get_classes():
        if is_ascii_problem(classe.get_name()):
            return True
        for method in classe.get_methods():
            if is_ascii_problem(method.get_name()):
                return True
    return False
